/*
 * Copyright (c) 2019, ams AG
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include "sns_tmf8801_sensor.h"
#include "sns_time.h"

static tof_deviceIdentifier_t tmf8801_deviceIdentifier[]={
    {AMS_DEVICE_ID, AMS_DEVICE_ID_MASK, 0,          AMS_REV_ID_MASK, AMS_TMF8801},
    {AMS_DEVICE_ID, AMS_DEVICE_ID_MASK, AMS_REV_ID, AMS_REV_ID_MASK, AMS_TMF8801},
    {0, 0, 0, 0, AMS_LAST_DEVICE}
};

static intelHexAddr_t intelHexAddr = {
    .ulba = 0,
    .addr = 0
};
static bool tofPowerUp = 1;
static ams_deviceCtx_t tmf8801_device_contexts[tmf8801_NUM_DEVICES];
static uint8_t tmf8801_clockSkew_correctResults(sns_sensor_instance *instance, ams_deviceCtx_t * ctx);
static uint8_t tmf8801_waitForCpuReady(AMS_PORT_portHndl * portHndl, void * ptr);
static void tmf8801_tofInit(ams_deviceCtx_t *dcbCtx);
static void tmf8801_cr_init(tof_clk_cr_t * cr);
static uint8_t tmf8801_BL_addrRam(AMS_PORT_portHndl * portHndl, uint32_t addr);
static uint8_t tmf8801_BL_writeRam(AMS_PORT_portHndl * portHndl, uint8_t * dat_in, uint8_t len);
static uint8_t tmf8801_BL_ramRemap(AMS_PORT_portHndl * portHndl);

#ifdef ENABLE_DEBUG
/**
 * Function to exit the island mode on demand
 * @param this
 */
static void inst_exit_island(sns_sensor_instance *this);
#endif

uint8_t AMS_tmf8801_rwBlock(AMS_PORT_portHndl * portHndl, uint8_t reg_addr, uint8_t * data, uint8_t length, bool isWrite) {

    uint32_t rw_count;
    sns_rc status;
    sns_port_vector port_vec;
    port_vec.buffer = data;
    port_vec.bytes = length;
    port_vec.is_write = isWrite;
    port_vec.reg_addr = reg_addr;

    if (portHndl == NULL)
        return 0;

    status = portHndl->scp_service->api->sns_scp_register_rw(portHndl->scp_handle,
                                                            &port_vec,
                                                            1,
                                                            false,
                                                            &rw_count);

    if (status == SNS_RC_SUCCESS) {
        return (uint8_t)rw_count;
    }

    return 0;
}

 uint8_t AMS_tmf8801_setByte(AMS_PORT_portHndl * portHndl, uint8_t reg_addr, uint8_t data)
{
    uint8_t write_count = 0;

    write_count = AMS_tmf8801_rwBlock(portHndl,reg_addr,&data,1, I2C_WRITE);

    return write_count;
}

uint8_t AMS_string_to_bytes(char* str_in, uint8_t str_length,
                                  uint8_t* data_out, uint8_t data_length)
{
    char* pos = str_in;
    int count;
    for (count = 0; count < data_length; count++) {
        sscanf(pos, "%2hhx", &data_out[count]);
        pos += 2;
        if((count<<1) >= str_length){
            return 0;
        }
    }
    return count;
}

uint8_t AMS_bytes_to_string(uint8_t* data, uint8_t data_length,
                             char* output, uint8_t output_length)
{
  // if not enough space was available, returns -1
  // otherwise returns the number of characters written to
  // output, not counting the additional null character

  uint8_t written = 0;
  uint8_t length;
  for (; data_length; data_length--) {
    length = snprintf(output, output_length, "%02x", *data++);
    if (length >= output_length) {
      // not enough space
      return 0;
    }
    written += length;
    output += length;
    output_length -= length;
  }
  return written;
}

static uint8_t tmf8801_write_command(AMS_PORT_portHndl * portHndl,uint8_t cmd, uint8_t *dat_buf, uint8_t num)
{
    uint8_t wc;
    // write parameters
    uint8_t buf[TMF8801_CMD_TOTAL_SIZE];
    uint8_t offset = TOF8801_REG_CMD_DAT9 + TMF8801_CMD_PARAMS_SIZE - num;

    sns_memscpy(buf, TMF8801_CMD_TOTAL_SIZE, dat_buf, num);
    buf[num] = cmd;
    wc = AMS_tmf8801_rwBlock(portHndl, offset, buf, (num+1), I2C_WRITE);

    return wc;
}


uint8_t tmf8801_download_systemData(AMS_PORT_portHndl * portHndl, tof_system_data_t * sysDat)
{

    uint8_t length;

    if((sysDat->calReceived) && (sysDat->stateReceived))
    {
        length = TMF8801_CALIBRATION_DATA_SIZE + TMF8801_ALG_STATE_DATA_SIZE;

    }
    else if(sysDat->calReceived)
    {
        length = TMF8801_CALIBRATION_DATA_SIZE;
    }
    else
    {
        return 0;
    }

    AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_DATA_SEC_START, (uint8_t*)sysDat, length, I2C_WRITE);

    return length;
}

static uint8_t tmf8801_setEnable(ams_deviceCtx_t * ctx)
{
    uint8_t dat;
    if (ctx->allowIens)
    {
        dat = TOF8801_MEASUREMENT_INT_EN;
    }
    else
    {
        dat = 0;
    }
    return AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_INT_EN,  dat);

}

uint8_t tmf8801_deviceCalibrate(sns_sensor_instance *instance, ams_deviceCtx_t * ctx){
    uint8_t wc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_deviceCalibrate: We are here");

    tmf8801_tofInit(ctx);
    wc = AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_CMD, TOF8801_CMD_START_CALIBRATE);
    if(!wc)
        return wc;
    wc = AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_INT_EN,  TOF8801_MEASUREMENT_INT_EN);
    ctx->tofState = PRX_STATE_FACTORY_CALIBRATION;
    ctx->enableCal = 0;
    // to prevent downloading state data and cal data that are read from registry
    tofPowerUp = false;
    return wc;
}


static uint8_t tmf8801_deviceStartCapture(sns_sensor_instance *instance, ams_deviceCtx_t * ctx){

    uint8_t wc = 1;
    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_deviceStartCapture: enter ...");

    if(ctx->tofState == PRX_STATE_IDLE)
    {
        uint8_t tempbuf[TOF8801_CMD_MEASURE_NO_DOWNLAOD_PARAMS_SIZE+1];
        uint8_t *tmpPtr = (uint8_t*)&ctx->ctxCmdData->alg;

        sns_memscpy(&tempbuf[1], TOF8801_CMD_MEASURE_NO_DOWNLAOD_PARAMS_SIZE, tmpPtr, TOF8801_CMD_MEASURE_NO_DOWNLAOD_PARAMS_SIZE);
        // download either or both or non of cal data / state data
        tempbuf[0] = ctx->ctxSystemData->calReceived | (ctx->ctxSystemData->stateReceived<<1);
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_deviceStartCapture: download cal-state setting is 0x%x", tempbuf[0]);
        wc =  tmf8801_write_command(ctx->portHndl,TOF8801_CMD_START_CAP_WITH_DOWNLAOD, tempbuf, (TOF8801_CMD_MEASURE_NO_DOWNLAOD_PARAMS_SIZE+1));
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_deviceStartCapture: wrote %d command data", wc);
        if(!wc)
            return wc;
        AMS_DEV_INST_PRINTF(LOW, instance,
        "AMS: tmf8801_deviceStartCapture: write_command_data iteration = 0x%x, period = %d, threshold = %d, delay = %d, gpio = 0x%x, alg = 0x%x\n",
        ctx->ctxCmdData->iteration[0]|(ctx->ctxCmdData->iteration[1]<<8),ctx->ctxCmdData->period,
        ctx->ctxCmdData->threshold, ctx->ctxCmdData->delay,ctx->ctxCmdData->gpio,ctx->ctxCmdData->alg);
        //wc = AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_INT_EN,  TOF8801_MEASUREMENT_INT_EN);
        wc = tmf8801_setEnable(ctx);
        AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_deviceStartCapture: enable_interrupt");
        ctx->tofState = PRX_STATE_CAPTURE;
    }
    return wc;
}

static uint8_t tmf8801_deviceResumeCapture(sns_sensor_instance *instance, ams_deviceCtx_t * ctx){

    uint8_t wc = 1;
    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_deviceResumeCapture: enter ....");
    if(ctx->tofState == PRX_STATE_IDLE)
    {
        wc =  tmf8801_write_command(ctx->portHndl,TOF8801_CMD_START_CAP_NO_DOWNLAOD, (uint8_t*)&ctx->ctxCmdData->alg, TOF8801_CMD_MEASURE_NO_DOWNLAOD_PARAMS_SIZE);
        if(!wc)
            return wc;
        AMS_DEV_INST_PRINTF(HIGH, instance,
        "AMS: tmf8801_deviceResumeCapture : write_command_data iteration = 0x%x, period = %d, threshold = %d, delay = %d, gpio = 0x%x, alg = 0x%x\n",
        ctx->ctxCmdData->iteration[0]|(ctx->ctxCmdData->iteration[1]<<8),ctx->ctxCmdData->period,
        ctx->ctxCmdData->threshold, ctx->ctxCmdData->delay,ctx->ctxCmdData->gpio,ctx->ctxCmdData->alg);
        //wc = AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_INT_EN,  TOF8801_MEASUREMENT_INT_EN);
        wc = tmf8801_setEnable(ctx);
        ctx->tofState = PRX_STATE_CAPTURE;
    }
    return wc;
}

static uint8_t tmf8801_deviceCapture(sns_sensor_instance *instance, ams_deviceCtx_t * ctx){

    uint8_t wc;
    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_deviceCapture: enter ....");
    if(tofPowerUp){
        wc = tmf8801_deviceStartCapture(instance, ctx);
        tofPowerUp = false;
    }
    else{
        wc = tmf8801_deviceResumeCapture(instance, ctx);
    }
    return wc;
}

static uint8_t tmf8801_deviceStop(sns_sensor_instance *instance, ams_deviceCtx_t * ctx){

    uint8_t wc;
    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_deviceStop: enter ..");
        //disable interrupt
    wc =  AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_INT_EN,  0);
    if(!wc)
        return wc;
        //stop measurement
    wc =  AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_CMD,  TOF8801_CMD_STOP_CAPTURE);
        //set state to idle
    ctx->tofState = PRX_STATE_IDLE;
    wc = tof8801_app0WaitForIdle(ctx->portHndl);

    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_deviceStop: check if device capture fully stops : %d", wc);
    return wc;
}

void tmf8801_hexDecode_writeRam(AMS_PORT_portHndl * portHndl, uint8_t* data_in)
{
    switch (data_in[INTEL_HEX_TYPE_INDEX])
    {
        case INTEL_HEX_TYPE_DATA:
        {
            intelHexAddr.addr = (data_in[INTEL_HEX_ADDR_HIGH_INDEX]<<8) | (data_in[INTEL_HEX_ADDR_LOW_INDEX]) | intelHexAddr.ulba;
            tmf8801_BL_addrRam(portHndl, intelHexAddr.addr);
            tmf8801_BL_writeRam(portHndl, &data_in[INTEL_HEX_DATA_0_INDEX], data_in[INTEL_HEX_BYTE_COUNT_INDEX]);
            break;
        }
        case INTEL_HEX_TYPE_EOF:
        {
            tmf8801_BL_ramRemap(portHndl);
            break;
        }
        case INTEL_HEX_TYPE_EXT_LIN_ADDR:
        {
            intelHexAddr.ulba = (data_in[INTEL_HEX_DATA_0_INDEX]<<23) | (data_in[INTEL_HEX_DATA_1_INDEX]<<16);
            break;
        }
        default:
        break;
    }
}

/**
 * tof8801_calc_chksum - calculate the chksum and return the result
 *
 * @data: pointer to data to chksum
 * @size: number of elements to chksum over in data
 */
uint8_t tmf8801_calc_chksum(const uint8_t *data, uint8_t size)
{
    uint32_t sum = 0;
    int32_t idx = 0;
    for (; idx < size; idx++) {
        sum += data[idx];
    }
    return (uint8_t) ~sum; /* 1's complement of lowest byte */
}

/**
 * tof8801_BL_read_status - Retry until cmd is executed and status is
 *                          available in local buffer
 *
 * @client: pointer to tmf8801_ctx
 * @BL_app: pointer to BL application struct
 * @num_retries: how many times to retry to retrieve status
 */


uint8_t tof8801_BL_readStatus(AMS_PORT_portHndl * portHndl)
{
    uint8_t error = 1;
    uint8_t bl_rbuf[BL_I2C_CMD_MAX_SIZE + 1];
    uint8_t *status = &bl_rbuf[0];
    uint8_t *rdata_size = &bl_rbuf[1];
    uint8_t chksum;
    int8_t num_retries = 5;
    do {
        num_retries -= 1;

        AMS_tmf8801_rwBlock(portHndl, TOF8801_BL_CMD_STAT, bl_rbuf, BL_I2C_HEADER_SIZE, I2C_READ);

        if (BL_IS_CMD_BUSY(*status)) {
            /* CMD is still executing, wait and retry */
            sns_busy_wait(usec_1000_in_ticks);
            if (num_retries <= 0) {
                error = 0;
            }
            continue;
        }
        /* if we have reached here, the command has either succeeded or failed */
        if ( *rdata_size > 0 ) {
            /* read in data part and csum */

            AMS_tmf8801_rwBlock(portHndl, TOF8801_BL_CMD_STAT, bl_rbuf, BL_CALC_BL_RSP_SIZE(*rdata_size), I2C_READ);

            chksum = (uint8_t) ~tmf8801_calc_chksum(bl_rbuf, BL_CALC_BL_RSP_SIZE(*rdata_size));
            if ((chksum != BL_VALID_CHKSUM) && (num_retries <= 0)) {
                return 0;
            }
            /* all done, break and return */
            break;
        }
    } while ((error == 1) && (num_retries > 0));
    return error;
}

uint8_t tmf8801_BL_sendRcvCmd(AMS_PORT_portHndl * portHndl, uint8_t* dat_buf, uint8_t num)
{
    uint8_t rc;
    rc = AMS_tmf8801_rwBlock(portHndl, TOF8801_BL_CMD_STAT, dat_buf, num, I2C_WRITE);

    if(!rc)
        return 0;

    return tof8801_BL_readStatus(portHndl);
}
uint8_t tmf8801_BL_uploadInit(AMS_PORT_portHndl * portHndl, uint8_t salt)
{
    tof8801_BL_upload_init_cmd_t cmd;
    cmd.command = BL_UPLOAD_INIT;
    cmd.size = 1;
    cmd.seed = salt;
    cmd.chksum = tmf8801_calc_chksum((uint8_t*)&cmd,BL_CALC_CHKSUM_SIZE(cmd.size));

    return tmf8801_BL_sendRcvCmd(portHndl, (uint8_t*)&cmd, BL_CALC_BL_CMD_SIZE(cmd.size));
}

static uint8_t tmf8801_BL_addrRam(AMS_PORT_portHndl * portHndl, uint32_t addr)
{
    tof8801_BL_addr_ram_cmd_t cmd;
    cmd.command = BL_ADDR_RAM;
    cmd.size = 2;
    cmd.addr_lsb  = addr & 0xff;
    cmd.addr_msb  = (addr >> 8) & 0xff;
    cmd.chksum = tmf8801_calc_chksum((uint8_t*)&cmd,BL_CALC_CHKSUM_SIZE(cmd.size));

    return tmf8801_BL_sendRcvCmd(portHndl, (uint8_t*)&cmd, BL_CALC_BL_CMD_SIZE(cmd.size));
}

static uint8_t tmf8801_BL_writeRam(AMS_PORT_portHndl * portHndl, uint8_t * dat_in, uint8_t len)
{
    tof8801_BL_write_ram_cmd_t cmd;
    int32_t idx = 0;

    cmd.command = BL_W_RAM;
    cmd.size = len;
    for(; idx < cmd.size; idx++) {
        cmd.data[idx] = dat_in[idx];
    }
    /* add chksum to end */
    cmd.data[(uint8_t)cmd.size] =
        tmf8801_calc_chksum((uint8_t*)&cmd,BL_CALC_CHKSUM_SIZE(cmd.size));
    return tmf8801_BL_sendRcvCmd(portHndl, (uint8_t*)&cmd, BL_CALC_BL_CMD_SIZE(cmd.size));
}

static uint8_t tmf8801_BL_ramRemap(AMS_PORT_portHndl * portHndl)
{
    tof8801_BL_short_cmd_t cmd;
    cmd.command = BL_RAMREMAP_RESET;
    cmd.size = 0;
    cmd.chksum = tmf8801_calc_chksum((uint8_t*)&cmd, BL_CALC_CHKSUM_SIZE(cmd.size));
    AMS_tmf8801_rwBlock(portHndl, TOF8801_BL_CMD_STAT, (uint8_t*)&cmd, BL_CALC_BL_CMD_SIZE(cmd.size), I2C_WRITE);

    return tmf8801_waitForCpuReady(portHndl, NULL);
}

static void tmf8801_getCalResult(sns_sensor_instance *const this, ams_deviceCtx_t* dCtx)
{
    ams_instance_state *state = (ams_instance_state*)this->state->state;
    state->instanceSystemData.calBlock = dCtx->ctxSystemData->calBlock;
    for(int i = 0; i < TMF8801_CALIBRATION_DATA_SIZE; i++)
    {
        state->instanceCalDataFloat.calData[i] = (float)dCtx->ctxSystemData->calBlock.calData[i];
    }
    //update the state data accordingly
    state->instanceSystemData.stateBlock.algStateData[0] =
            (dCtx->ctxSystemData->calBlock.calData[0]>>4) | (dCtx->ctxSystemData->calBlock.calData[1]<<4);
    state->instanceSystemData.stateBlock.algStateData[1] =
            (dCtx->ctxSystemData->calBlock.calData[1]>>4) | (dCtx->ctxSystemData->calBlock.calData[2]<<4);
    state->instanceSystemData.stateBlock.algStateData[2] = (dCtx->ctxSystemData->calBlock.calData[2]>>4);
    sns_memset((uint8_t*)&state->instanceSystemData.stateBlock.algStateData[3], 0, (TMF8801_ALG_STATE_DATA_SIZE-3));
}

static void tmf8801_getStateResult(sns_sensor_instance *const this, ams_deviceCtx_t* dCtx)
{
    ams_instance_state *state = (ams_instance_state*)this->state->state;

    sns_memscpy((uint8_t*)state->instanceSystemData.stateBlock.algStateData,3,
                (uint8_t*)dCtx->tofCapData.capStateData, 3);
    sns_memset((uint8_t*)&state->instanceSystemData.stateBlock.algStateData[3], 0, (TMF8801_ALG_STATE_DATA_SIZE-3));
}

/*
* In SEE interrupts should only be enabled after receiving
* SNS_INTERRUPT_MSGID_SNS_INTERRUPT_REG_EVENT.
*
* Therefore, if SNS_INTERRUPT_MSGID_SNS_INTERRUPT_REG_EVENT has been
* received in ams_inst_notify_event() then
*   ...
*   tmf8801_deviceSetConfig(state->priv, AMS_CONFIG_GLOBAL, AMS_CONFIG_ALLOW_INTS, 1);
*   ...
*
* When there are no more SEE clients _tmf8801_setup_clean_irq_stream() will called and
* it will call
*   ...
*   tmf8801_deviceSetConfig(state->priv, AMS_CONFIG_GLOBAL, AMS_CONFIG_ALLOW_INTS, 0);
*   ...
*
*/

bool tmf8801_deviceSetConfig(sns_sensor_instance *const this, ams_deviceCtx_t * ctx, \
                             ams_configureFeature_t feature, deviceConfigOptions_t option, uint32_t data){
#ifdef AMS_DEV_DEBUG_ENABLE
    {
        uint8_t tempbyte;
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_deviceSetConfig: tofState is %d\n", ctx->tofState);
        AMS_tmf8801_rwBlock(ctx->portHndl, TOF8801_REG_INT_EN, &tempbyte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_deviceSetConfig: device_interrupt_enabled is 0x%x\n", tempbyte);
    }
#endif

    AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceSetConfig: enter ......\n");
    if (feature == AMS_CONFIG_GLOBAL){
        AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceSetConfig: ams_configureFeature_t  AMS_CONFIG_GLOBAL(%d)\n", data);
        switch (option)
        {
            case AMS_CONFIG_ALLOW_INTS:
                if (data == 0) {
                    ctx->allowIens = false;
                } else {
                    ctx->allowIens = true;
                }
                tmf8801_setEnable(ctx);
                break;

            default:
                AMS_INST_PRINTF(ERROR, this, "AMS: tmf8801_deviceSetConfig _GLOBAL: ERROR, unk. option %d\n", option);
                return true;
        }
    }

    if (feature == AMS_CONFIG_PROX){
        AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceSetConfig: ams_configureFeature_t  AMS_CONFIG_PROX\n");
        switch (option)
        {
            case AMS_CONFIG_ENABLE: /* power on */
                AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceSetConfig: deviceConfigOptions_t AMS_CONFIG_ENABLE(%u), current mode %d", data, ctx->tofState);
                if (data == 0) {
                    ctx->enableCapture = false;
                } else {
                    /*  this will reinitialize the
                        proximity in a way that will force interrupt
                        before we were missing this interrupt because when initialliy other
                        feature was enabled the interrupt was generated but it was missed
                        since the interrupt was not configured because it wasn't needed
                    **/
                    ctx->enableCapture = true;
                }
                break;
            case AMS_CONFIG_ALWAYS_READ:
                AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceSetConfig: AMS_CONFIG_ALWAYS_READ(%d)\n", data);
                ctx->alwaysReadProx = !!data;
                if(ctx->alwaysReadProx){
                    ctx->enableCapture = true;
                }
                else{
                    ctx->enableCapture = false;
                }
                break;
            default:
                break;
        }
        if(ctx->enableCapture)
        {
            tmf8801_deviceCapture(this, ctx);
        }
        else
        {
            tmf8801_deviceStop(this, ctx);
        }
    }
#ifdef AMS_DEV_DEBUG_ENABLE
    {
        uint8_t tempbyte;
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_deviceSetConfig: tofState is %d\n", ctx->tofState);
        AMS_tmf8801_rwBlock(ctx->portHndl, TOF8801_REG_INT_EN, &tempbyte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_deviceSetConfig: device interrupt enabled is 0x%x\n", tempbyte);
    }
#endif

    AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceSetConfig : done .....\n");
    return 0;
}

bool tmf8801_deviceEventHandler(sns_sensor_instance *const this, ams_deviceCtx_t * ctx){

    int ret = 0;
    uint8_t int_st;
    uint8_t w_count;
    uint8_t buf[TMF8801_CAPTURE_TIMESTAMP_BLOCK_SIZE];

#ifdef ENABLE_DEBUG
    inst_exit_island((sns_sensor_instance *)this);
#endif
    AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceEventHandler: enter...\n");
#ifdef AMS_DEV_DEBUG_ENABLE
    {
        uint8_t tempbyte;
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_deviceEventHandler: tofState is %d\n", ctx->tofState);
        AMS_tmf8801_rwBlock(ctx->portHndl, TOF8801_REG_INT_EN, &tempbyte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_deviceEventHandler: device interrupt enabled is 0x%x\n", tempbyte);
    }
#endif

    //read interrupt status
    AMS_tmf8801_rwBlock(ctx->portHndl, TOF8801_REG_INT_STAT, &int_st, 1, I2C_READ);
    AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceEventHandler: int_status is 0x%x\n", int_st);
    //clear the intrrupt as soon as possible
    if(int_st != 0)
    {
        AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_INT_STAT, 0xFF);
    }
    if(int_st & TOF8801_MEASUREMENT_INT_EN)
    {
        /*********************************************************************/
        switch (ctx->tofState)
        {
            case PRX_STATE_FACTORY_CALIBRATION:
            {

                AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceEventHandler: PRX_STATE_FACTORY_CALIBRATION\n");

                // read out the calibration data
                w_count = AMS_tmf8801_rwBlock(ctx->portHndl, TOF8801_REG_DATA_SEC_START, (uint8_t*)&(ctx->ctxSystemData->calBlock.calData[0]), TMF8801_CALIBRATION_DATA_SIZE, I2C_READ);
                AMS_DEV_INST_PRINTF(HIGH, this,
                                    "AMS: tmf8801_deviceEventHandler: read_calibration_data total_words_read(%d) 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",
                                    w_count,
                                    ctx->ctxSystemData->calBlock.calData[0],ctx->ctxSystemData->calBlock.calData[1],ctx->ctxSystemData->calBlock.calData[2],
                                    ctx->ctxSystemData->calBlock.calData[3],ctx->ctxSystemData->calBlock.calData[4],ctx->ctxSystemData->calBlock.calData[5],
                                    ctx->ctxSystemData->calBlock.calData[6]);

                AMS_DEV_INST_PRINTF(HIGH, this,
                                    "AMS: tmf8801_deviceEventHandler: read_calibration_data 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",
                                    ctx->ctxSystemData->calBlock.calData[7],ctx->ctxSystemData->calBlock.calData[8],
                                    ctx->ctxSystemData->calBlock.calData[9],ctx->ctxSystemData->calBlock.calData[10],ctx->ctxSystemData->calBlock.calData[11],
                                    ctx->ctxSystemData->calBlock.calData[12],ctx->ctxSystemData->calBlock.calData[13]);
                ctx->updateAvailable = AMS_CALIBRATION_DONE;
                ctx->updateAvailable |= (1 << AMS_PROXIMITY_SENSOR);
                tmf8801_deviceStop(this, ctx);
                tmf8801_getCalResult(this,ctx);
                AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceEventHandler: updateAvailable = 0x%x", ctx->updateAvailable);
                break;
            }
            case PRX_STATE_CAPTURE:
            {

                AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_deviceEventHandler: PRX_STATE_CAPTURE \n");
                w_count = AMS_tmf8801_rwBlock(ctx->portHndl, TOF8801_REG_STATUS,
                                    buf, TMF8801_CAPTURE_TIMESTAMP_BLOCK_SIZE, I2C_READ);
                ctx->sns_usec_stamp = sns_get_system_time();
                sns_memscpy((uint8_t*)&ctx->tofCapData,TMF8801_CAPTURE_BLOCK_SIZE, (uint8_t*)&buf[3], TMF8801_CAPTURE_BLOCK_SIZE);

                AMS_DEV_INST_PRINTF(HIGH, this,
                "AMS: tmf8801_deviceEventHandler: read_capture_data total_words_read(%d) %d, 0x%x, %d, 0x%x, 0x%x, 0x%x, 0x%x\n",
                                    w_count,
                                    ctx->tofCapData.capIndex,ctx->tofCapData.reliability,ctx->tofCapData.distance,
                                    ctx->tofCapData.timeStamp,
                                    ctx->tofCapData.capStateData[0],ctx->tofCapData.capStateData[1],ctx->tofCapData.capStateData[2]);
                AMS_DEV_INST_PRINTF(HIGH, this,
                "AMS: tmf8801_deviceEventHandler: read_state_data 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",
                                    ctx->tofCapData.capStateData[3],ctx->tofCapData.capStateData[4],ctx->tofCapData.capStateData[5],
                                    ctx->tofCapData.capStateData[6],ctx->tofCapData.capStateData[7],ctx->tofCapData.capStateData[8],
                                    ctx->tofCapData.capStateData[9],ctx->tofCapData.capStateData[10]);
                tmf8801_clockSkew_correctResults(this, ctx);
                ctx->updateAvailable |= (1 << AMS_PROXIMITY_SENSOR);
                tmf8801_getStateResult(this, ctx);
                AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_deviceEventHandler: updateAvailable = 0x%x", ctx->updateAvailable);
                break;
            }
            default:
                break;
        }
    }
    return ret;
}

uint32_t tmf8801_getUpdateResult(ams_deviceCtx_t * ctx) {
    uint32_t returnValue = ctx->updateAvailable;
    ctx->updateAvailable = 0;
    return returnValue;
}

uint8_t tmf8801_switch_app(AMS_PORT_portHndl * portHndl, uint8_t app_id){
  uint8_t error;
  error = AMS_tmf8801_setByte(portHndl, TOF8801_REG_REQ_APP_ID, app_id);

  return error;
}

void * amsLibrary_open_tmf8801(AMS_PORT_portHndl* port_handle_ptr, tof_system_data_t * sysDat, \
                               tof_cmd_data_t * cmdDat, ams_deviceCtx_t *amsLibrary, int32_t hwid){
    tof_deviceIdentifier_e chipId = AMS_UNKNOWN_DEVICE;
    ams_deviceCtx_t *p_devctx = &amsLibrary[hwid];;
    UNUSED_VAR(sysDat);
    UNUSED_VAR(cmdDat);
    if (NULL == amsLibrary) {
        chipId = tmf8801_validateDevice(port_handle_ptr);
        if (chipId != AMS_UNKNOWN_DEVICE){
            p_devctx = &tmf8801_device_contexts[hwid];
        }
        else{
            p_devctx = NULL;
        }
    }
    else{
        p_devctx = NULL;
    }
    return (void*)p_devctx;
}


uint8_t tmf8801_deviceInit(ams_deviceCtx_t * ctx, AMS_PORT_portHndl * portHndl, tof_system_data_t * sysDat, \
                         tof_cmd_data_t * cmdDat, tof_deviceIdentifier_e chipId, int32_t hwid, uint8_t* fw_ver){

    sns_memset((void*)ctx, 0, sizeof(ams_deviceCtx_t));
    ctx->hwid = hwid;
    ctx->portHndl = portHndl;
    ctx->ctxSystemData = sysDat;
    ctx->ctxCmdData = cmdDat;
    ctx->deviceId = chipId;
    sns_memscpy(ctx->tof_fwVersion, 5, fw_ver, 5);
    ctx->tofState = PRX_STATE_IDLE;
    tmf8801_tofInit(ctx);
    return 0;
}

uint8_t tmf8801_devicePowerUp(AMS_PORT_portHndl * portHndl, sns_sensor *this){
    uint8_t ret;
    ret = tmf8801_waitForCpuReady(portHndl, this);
    tmf8801_BL_uploadInit(portHndl, BL_DEFAULT_SALT);
    return ret;
}

tof_deviceIdentifier_e tmf8801_validateDevice(AMS_PORT_portHndl * portHndl){

    uint8_t chipId;

    if( portHndl == NULL)
        return 0;

    AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_DVICE_ID, &chipId, 1, I2C_READ);
    if(TOF8801_DEVICE_IS_VALID(chipId)){
        return tmf8801_deviceIdentifier[0].device;
    }
    else{
        return AMS_UNKNOWN_DEVICE;
    }
}


static uint8_t tmf8801_waitForCpuReady(AMS_PORT_portHndl * portHndl, void * ptr){

  int retry = 0;
  int error = 0;
  uint8_t status;
  uint8_t wr_en = 0;
  sns_sensor * sensor = NULL;
  if( ptr!=NULL){
      wr_en = 1;
      sensor = (sns_sensor *) ptr;
  }
  while (retry++ < TOF8801_MAX_STARTUP_RETRY) {
      error = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_ENABLE, &status, 1, I2C_READ);
      if (TOF8801_STAT_CPU_READY(status)) {
          if(wr_en){
          AMS_DEV_PRINTF(LOW, sensor, "AMS: tmf8801_waitForCpuReady: CPU ready == 0x41 ");
          }
          return 1;
      } else if (TOF8801_STAT_CPU_SLEEP(status)) {
          AMS_tmf8801_setByte(portHndl, TOF8801_REG_ENABLE, CPU_WAKEUP);
          sns_busy_wait(usec_100_in_ticks<<1);
      }
  }
  if(wr_en){
    AMS_PRINTF(ERROR, sensor, "AMS: tmf8801_waitForCpuReady: CPU wake up timeout with failure");
  }
  return error;
}

uint8_t tmf8801_cpuGoStandby(AMS_PORT_portHndl * portHndl)
{
    uint8_t wc;
    uint8_t status;
    uint32_t usec_sleep = 5000;
    uint32_t usec_timeout;
    uint32_t total_sleep = 0;
    usec_timeout = usec_sleep*10;
    wc = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_ENABLE, &status, 1, I2C_READ);
    if (wc && !TOF8801_STAT_CPU_SLEEP(status)) {
        // set PON = 0 to go to sleep
        status &= ~CPU_PON0_MASK;
        wc = AMS_tmf8801_setByte(portHndl, TOF8801_REG_ENABLE, status);
        do {
            sns_busy_wait(usec_5000_in_ticks);
            wc = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_ENABLE, &status, 1, I2C_READ);
            if ((wc) && (TOF8801_STAT_CPU_STANDBY(status))){
                do {
                    wc = AMS_tmf8801_rwBlock(portHndl, 0xEC, &status, 1, I2C_READ);
                    if ((wc) && ((status&0x40)==0)){
                        return 1;
                    }
                    sns_busy_wait(usec_100_in_ticks<<1);
                    total_sleep += 200;
                 }while (total_sleep < usec_timeout);
            }
            total_sleep += usec_sleep;
        } while (total_sleep < usec_timeout);
    }
    return 0;
}

/*  simplified version
uint8_t tmf8801_cpuGoStandby(AMS_PORT_portHndl * portHndl)
{
    uint8_t wc;
    uint8_t status;

    wc = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_ENABLE, &status, 1, I2C_READ);
    if (wc && !TOF8801_STAT_CPU_SLEEP(status)) {
        // set PON = 0 to go to sleep
        status &= ~CPU_PON0_MASK;
        wc = AMS_tmf8801_setByte(portHndl, TOF8801_REG_ENABLE, status);
        sns_busy_wait(usec_100_in_ticks<<1);
    }
    return wc;
}
*/
uint8_t tmf8801_wakeFromStandby(AMS_PORT_portHndl * portHndl)
{
    uint8_t wc;
    uint8_t status;

    wc = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_ENABLE, &status, 1, I2C_READ);
    if (wc && TOF8801_STAT_CPU_SLEEP(status)) {
        // set PON = 0 to go to sleep
        status |= CPU_PON0_MASK;
        wc = AMS_tmf8801_setByte(portHndl, TOF8801_REG_ENABLE, status);
        sns_busy_wait(usec_100_in_ticks<<1);
    }
    return wc;
}

uint8_t tof8801_app0WaitForIdle(AMS_PORT_portHndl * portHndl)
{
    int32_t wc = 0;
    uint32_t usec_sleep = 10000;
    uint32_t usec_timeout;
    uint32_t total_sleep = 0;
    uint8_t status;
    uint8_t cpu_reg = 0;
    usec_timeout = usec_sleep*30;
    do {
        sns_busy_wait(usec_10000_in_ticks);
        wc = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_ENABLE, &status, 1, I2C_READ);
        if (wc) {
            wc = AMS_tmf8801_rwBlock(portHndl, TOF8801_REG_STATE, &cpu_reg, 1, I2C_READ);
            if ((wc) &&
                (TOF8801_STAT_CPU_READY(status)) &&
                (cpu_reg == APP0_IDLE_STATE)) {
                return 1;
            }
        }
        total_sleep += usec_sleep;
    } while (total_sleep < usec_timeout);
    return 0;
}

bool tmf8801_getDeviceInfo(ams_deviceInfo_t * info, int32_t hwid){

    sns_memset(info, 0, sizeof(ams_deviceInfo_t));

    info->numberOfSubSensors = hwid;
    info->deviceName  = "TOF";
    sns_memscpy(info->deviceModel, sizeof(info->deviceModel),
              STRINGIFY(CONFIG_AMS_PART_NAME), sizeof(info->deviceModel));
    info->tableSubSensors[info->numberOfSubSensors]= AMS_PROXIMITY_SENSOR;
    info->numberOfSubSensors++;
    info->tofSensor.driverName = "TOF";
    info->tofSensor.adcBits = 16;
    info->tofSensor.maxPolRate = 50;
    info->tofSensor.activeCurrent_uA = 33000;
    info->tofSensor.standbyCurrent_uA = 85;
    info->tofSensor.rangeMax = 2500;
    info->tofSensor.rangeMin = 0;
    info->tofSensor.resolution = 1.0;
    return true;
}

static void tmf8801_tofInit(ams_deviceCtx_t *ctx){
    ctx->clk_trim_enable = true;
    tmf8801_cr_init(&ctx->ctxClkCr);
}

#ifdef ENABLE_DEBUG
static void inst_exit_island(sns_sensor_instance *this)
{
    sns_service_manager *smgr = this->cb->get_service_manager(this);
    sns_island_service *island_service =
            (sns_island_service *)smgr->get_service(smgr, SNS_ISLAND_SERVICE);

    island_service->api->sensor_instance_island_exit(island_service, this);
}
#endif

/*
 *  Clock correction and clock trim
 *  Note: these functions are extracted from MCU driver, originally designed by ROwen
 */

static void tmf8801_cr_init(tof_clk_cr_t * cr)
{
    //
    // start a new recalculation cycle, beginning with the default ratio 0.2
    //
    cr->first_host = 0;
    cr->last_host  = 0;
    cr->first_i2c  = 0;
    cr->last_i2c   = 0;
    cr->ratioQ15   = APP0_EXP_FREQ_RATIO_Q15;
    cr->count      = 0;
    cr->trim_count = 0;
}

static void tmf8801_cr_addPair(sns_sensor_instance *this, tof_clk_cr_t * cr, uint32_t host, uint32_t i2c)
{
  uint32_t num = 0;
  uint32_t den = 0;
  uint32_t denQM15 = 0;
  uint32_t res;
  //
  // add a pair of host and I2C times
  //
  AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: enter....0x%x,  0x%x  \n", host, i2c);
  if (i2c == 0)
  {
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: return I2C is 0 \n");
    return;
  }
  if (host < cr->last_host)
  {
    //this should not be possible since host count reference is unix_epoch
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: host < cr->last_host \n");
    cr->count = 0;
  }
  if (i2c < cr->last_i2c)
  {
    //wraparound case
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: i2c < cr->last_i2c \n");
    cr->count = 0;
  }
  if (cr->count == 0)
  {
    cr->first_host = host;
    cr->first_i2c  = i2c;
  }
  cr->last_host  =  host;
  cr->last_i2c   =  i2c;
  cr->count      += 1;
  if (cr->count >= APP0_RATIO_MINCOUNT)
  {
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: cr->count >= APP0_RATIO_MINCOUNT \n");
    cr->trim_count += 1;
    num = cr->last_host - cr->first_host;
    den = cr->last_i2c  - cr->first_i2c;
    res = den*1000/num;
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: before shift num = %d,  den = %d  num/den = %d \n", num, den, res);
    while ((num < APP0_CR_WRAPAROUND_DIV2) && (den < APP0_CR_WRAPAROUND_DIV2))
    {
        num <<= 1;
        den <<= 1;
    };
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: after shift num = 0x%x,  den = 0x%x  \n", num, den);
    // round up, always positive
    denQM15 = ((den + (1 << 14)) >> 15);
    AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: after shift num = %d,  denQM15 = %d  ratio = %d\n", num, denQM15, (uint32_t)(num/denQM15));
    if ((denQM15 == 0) || (num > den)) {
      cr->count = 0;
      AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: denQM15 == 0) || (num > den \n");
    } else {
      cr->ratioQ15 = (num + (denQM15 >> 1))/denQM15;
      AMS_DEV_INST_PRINTF(MED, this, "AMS: tmf8801_cr_addPair: Ratio is updated ratio is %d \n", cr->ratioQ15);
    }
  }
}

static void tmf8801_cr_reCalc(tof_clk_cr_t * cr)
{
  //
  // start a new recalculation cycle, but don't lose the latest ratioQ15.
  //
  cr->count = 0;
}

static uint32_t tmf8801_cr_map(tof_clk_cr_t *cr, uint32_t distance)
{
  //
  // apply the mapping function to calculate a clock-corrected distance
  //
  return (distance * APP0_EXPECTEDRATIO * cr->ratioQ15 + (1 << 14)) >> 15;
}

static uint8_t tof8801_oscTrim(sns_sensor_instance *instance,ams_deviceCtx_t *ctx, int32_t change)
{
    uint8_t rc = 0;
    int32_t trim_val = 0;
    uint8_t fuse3 = 0;
    uint8_t fuse6 = 0;

    /*
       Oscillator Trimming procedure:
       1. write 0x06: 0x29
       2. write 0xE0 = 0x00 (PON0)
       3. read/write 0x03
       4. read/write 0x06 bit 6
       5. write 0xE0 = 0x01 (PON1)
       6. wait for CPU_READY, error if still 0x00 after timeout (chip reset)
       */
    rc = AMS_tmf8801_setByte(ctx->portHndl, TOF8801_REG_CMD_DAT9, TOF8801_CMD_OSC_TRIM);
    if (!rc)
        return rc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: sent to OSC_TRIM mode");

    rc = tmf8801_cpuGoStandby(ctx->portHndl);
    if (!rc)
        return rc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: cpu standby mode");

    //let standby mode stabilize, this is called when use the simple version of tmf8801_cpuGoStandby
    // let standby mode stabilize
    //sns_busy_wait(usec_5000_in_ticks<<1);
    /*** Start osc trimming ***/
    rc = AMS_tmf8801_rwBlock(ctx->portHndl, 0x03, &fuse3, 1, I2C_READ);
    if (!rc)
        return rc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: read fuse3 = 0x%x", fuse3);
    rc = AMS_tmf8801_rwBlock(ctx->portHndl, 0x06, &fuse6, 1, I2C_READ);
    if (!rc)
        return rc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: read fuse6 = 0x%x", fuse6);

    trim_val = (((int32_t)fuse3) << 1) | ((fuse6 & (0x01 << 6)) >> 6);
    trim_val += change;
    trim_val &= 0x1FF;
    fuse6 &= ~(0x01 << 6);
    fuse6 |= ((trim_val & 0x01) << 6);
    fuse3  = ((trim_val >> 1) & 0xFF);

    rc = AMS_tmf8801_setByte(ctx->portHndl, 0x03, fuse3);
    if (!rc)
        return rc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: write fuse3 = 0x%x", fuse3);
    rc = AMS_tmf8801_setByte(ctx->portHndl, 0x06, fuse6);
    if (!rc)
        return rc;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: write fuse6 = 0x%x", fuse6);
    /*** End osc trimming ***/

    rc = tmf8801_wakeFromStandby(ctx->portHndl);
    if (!rc)
        return rc;

    rc = tof8801_app0WaitForIdle(ctx->portHndl);
    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tof8801_oscTrim: check if device capture fully stops : %d", rc);
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tof8801_oscTrim: leaving");
    return rc;
}

static uint8_t tof8801_oscTrim_Operation(sns_sensor_instance *instance, ams_deviceCtx_t * ctx, int32_t dir)
{
    int32_t rc;

    rc = tmf8801_deviceStop(instance, ctx);
    if (!rc)
        return rc;

    rc =  tof8801_oscTrim(instance, ctx, dir);
    if (!rc)
        return rc;

    rc = tmf8801_deviceResumeCapture(instance, ctx);
    return rc;
}

static uint8_t tmf8801_oscTrim_check(sns_sensor_instance *instance, ams_deviceCtx_t * ctx)
{
    uint8_t rc = 0;
    uint32_t ratioQ15;
    static bool stable = false;
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_oscTrim_check: enter.... trim_count = %d", ctx->ctxClkCr.trim_count);
    if (ctx->ctxClkCr.trim_count >= APP0_DEFAULT_CLK_ITERATIONS) {

        ratioQ15 = ctx->ctxClkCr.ratioQ15;
        ratioQ15 = ratioQ15 < APP0_MIN_FREQ_RATIO_Q15 ? APP0_MIN_FREQ_RATIO_Q15 : ratioQ15;
        ratioQ15 = ratioQ15 > APP0_MAX_FREQ_RATIO_Q15 ? APP0_MAX_FREQ_RATIO_Q15 : ratioQ15;
        ctx->ctxClkCr.trim_count = 0;

        if (ctx->clk_trim_enable) {
            if (ratioQ15 >= APP0_FREQ_RATIO_HITH_Q15) {
                rc = tof8801_oscTrim_Operation(instance, ctx, AMS_OSC_TRIM_UP);
                stable = false;
                AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_oscTrim_check: AMS_OSC_TRIM_UP");
            } else if (ratioQ15 <= APP0_FREQ_RATIO_LOTH_Q15) {
                stable = false;
                rc = tof8801_oscTrim_Operation(instance, ctx, AMS_OSC_TRIM_DOWN);
                AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_oscTrim_check: AMS_OSC_TRIM_DOWN");
            } else if (stable == 0) {
                stable = true;
            }
        }
    }
    return rc;
}

static uint8_t tmf8801_clockSkew_correctResults(sns_sensor_instance *instance, ams_deviceCtx_t * ctx)
{
    uint8_t error = 0;
    uint64_t loc_usec_stamp;
    uint32_t usec_stamp;
    static uint32_t last_stamp = 0;
    uint32_t old_dist = 0;
    uint32_t cr_dist = 0;
    uint32_t tof_clk_cnt;

    // in uSec
    loc_usec_stamp = ((uint64_t)ctx->sns_usec_stamp*(uint64_t)sysTimeRes)/1000;
    usec_stamp = (uint32_t)(loc_usec_stamp & 0xFFFFFFFF);
    if ( abs((int32_t)(usec_stamp - last_stamp)) >= 60000000 ) {
        //reset the integration once/min
        tmf8801_cr_reCalc(&ctx->ctxClkCr);
        last_stamp = usec_stamp;
    }
    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_clockSkew_correctResults: sysTimeRes: 0x%x ns", (uint32_t)sysTimeRes);
    tof_clk_cnt = ctx->tofCapData.timeStamp;
    old_dist = ctx->tofCapData.distance;
    tmf8801_cr_addPair(instance, &ctx->ctxClkCr, usec_stamp, tof_clk_cnt);

    cr_dist = tmf8801_cr_map(&ctx->ctxClkCr, old_dist);
    if (cr_dist > APP0_MAX_DIST_MM) {
        // if corrected distance is out of range, remove the object detection
        cr_dist = 0;
        ctx->tofCapData.reliability &= ~RELIABILITY_MASK;
    }

    ctx->tofCapData.distance = cr_dist;

    AMS_DEV_INST_PRINTF(LOW, instance, "AMS: tmf8801_clockSkew_correctResults: clk_skew: host_ts: 0x%x us dev_ts: 0x%x(5MHz cnt) "
            "old_dist: %d new_dist: %d ratioQ15: %d, reliability: %d\n",
            usec_stamp, tof_clk_cnt, old_dist, cr_dist, ctx->ctxClkCr.ratioQ15, (ctx->tofCapData.reliability&RELIABILITY_MASK));

    error = tmf8801_oscTrim_check(instance, ctx);
    return error;
}

