/*
 * Copyright (c) 2019, ams AG
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include "sns_time.h"
#include "sns_tmf8801_sensor.h"

#define STRING_PB_SIZE_MAX   50

static bool ram_downloaded = 0;

tof_cmd_data_t  default_tof_cmd_data = {
    .iteration = {TMF8801_CMD_ITERATION_DEFAULT, TMF8801_CMD_ITERATION_DEFAULT},
    .threshold = TMF8801_CMD_THRESHOLD_DEFAULT,
    .period    = TMF8801_CMD_PERIOD_DEFAULT,
    .delay     = TMF8801_CMD_DELAY_DEFAULT,
    .gpio      = TMF8801_CMD_GPIO_DEFAULT,
    .alg       = TMF8801_CMD_ALG_DEFAULT
};

static char *registry_platform_placement_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME("_platform.placement")};
static char *registry_platform_config_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME("_platform.config")};
static char *registry_tof_fac_cal_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME("_platform.tof.fac_cal")};
static char *registry_tof_cmd_data_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME("_platform.tof.cmd_data")};
static char *registry_tof_state_data_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME("_platform.tof.state_data")};
static char *registry_tof_config_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME(".tof.config")};
static char *registry_tof_ram_names[tmf8801_NUM_DEVICES] = { AMS_REGISTRY_ENTRY_NAME("_0_firmware.fw")};

//define buffer for ram download
static char global_parsed_char_array[STRING_PB_SIZE_MAX] = {'0'};
static uint32_t ram_rec_size_total = 1;
static uint8_t int_buf[25] = {0};

static void tmf8801_send_fifo_flush_done(sns_sensor_instance *const instance,
                              sns_sensor_uid *suid)
{
  sns_service_manager *mgr = instance->cb->get_service_manager(instance);
  sns_event_service *e_service = (sns_event_service*)mgr->get_service(mgr, SNS_EVENT_SERVICE);
  sns_sensor_event *event = e_service->api->alloc_event(e_service, instance, 0);

  if(NULL != event)
  {
    event->message_id = SNS_STD_MSGID_SNS_STD_FLUSH_EVENT;
    event->event_len = 0;
    event->timestamp = sns_get_system_time();

    e_service->api->publish_event(e_service, instance, event, suid);
  }
}

/**
 * Returns decoded request message for type
 * sns_std_sensor_config.
 *
 * @param[in] in_request   Request as sotred in client_requests
 *                         list.
 * @param decoded_request  Standard decoded message.
 * @param decoded_payload  Decoded stream request payload.
 *
 * @return bool true if decode is successful else false
 */

static bool ams_get_decoded_std_request(sns_sensor const *this, sns_request const *in_request,
                                            sns_std_request *decoded_request,
                                            sns_std_sensor_config *decoded_payload)
{
  pb_istream_t stream;
  pb_simple_cb_arg arg =
      { .decoded_struct = decoded_payload,
        .fields = sns_std_sensor_config_fields };
  decoded_request->payload = (struct pb_callback_s)
      { .funcs.decode = &pb_decode_simple_cb, .arg = &arg };
  stream = pb_istream_from_buffer(in_request->request,
                                  in_request->request_len);
  if(!pb_decode(&stream, sns_std_request_fields, decoded_request))
  {
    AMS_PRINTF(ERROR, this, "AMS: ams_get_decoded_std_request: pb_decode error ");
    return false;
  }
  return true;
}


static void ams_get_device_config(sns_sensor *this,
                                   sns_sensor_instance *instance,
                                   ams_sensor_type sensor_type,
                                   float *chosen_sample_rate,
                                   float *chosen_report_rate,
                                   bool *client_present,
                                   int32_t *num_clients)
{
  sns_sensor_uid suid;
  sns_request const *request;
  ams_state *state = (ams_state*)this->state->state;
  sns_memscpy(&suid, sizeof(suid), &((sns_sensor_uid){{0}}), sizeof(sns_sensor_uid));
  sns_sensor_uid * p_suid;
  int32_t hwid = this->cb->get_registration_index(this);

  *num_clients = 0;
  *client_present = false;

  AMS_DEV_PRINTF(MED, this, "AMS: ams_get_device_config enter.........");

  if ((sensor_type == AMS_PROX_OC)   ||
      (sensor_type == AMS_PROX_STRM))
  {
      p_suid = (sns_sensor_uid *)tmf8801_get_suid_with_sensortype_and_hwid(sensor_type, hwid);
      sns_memscpy(&suid, sizeof(suid), p_suid, sizeof(sns_sensor_uid));
      AMS_DEV_PRINTF(MED, this, "AMS: ams_get_device_config get SUID: this %p, instance %p, sensor_type %d, "
                      , this
                      , instance
                      , sensor_type
                      );
  }
  else
  {
    return;
  }

  if (instance->cb->get_client_request(instance, &suid, true) != NULL)
  {
    *chosen_report_rate = 0;
    *chosen_sample_rate = 0;
  }
  /** Parse through existing requests and get fastest sample
   *  rate and report rate requests. */
  for(request = instance->cb->get_client_request(instance, &suid, true);
      NULL != request;
      request = instance->cb->get_client_request(instance, &suid, false))
  {
    sns_std_request decoded_request;
    sns_std_sensor_config decoded_payload = {0};

    AMS_DEV_PRINTF(MED, this, "AMS: ams_get_device_config: this %p, instance %p, sensor_type %d, msgId=%d"
                      , this
                      , instance
                      , sensor_type
                      , request->message_id
                      );

    if(request->message_id == SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_CONFIG)
    {
      if(ams_get_decoded_std_request(this, request, &decoded_request, &decoded_payload))
      {
        AMS_DEV_PRINTF(MED, this, "AMS: ams_get_device_config: sensor_type %d, curr SR %d, recvd SR %d",
                                sensor_type,
                                (int32_t)*chosen_sample_rate,
                                (int32_t)decoded_payload.sample_rate);
        *chosen_sample_rate = SNS_MAX(*chosen_sample_rate,
                                       decoded_payload.sample_rate);
        *chosen_report_rate = *chosen_sample_rate;
        *client_present = true;
        (*num_clients)++;
      }
    }
    else if(request->message_id == SNS_STD_SENSOR_MSGID_SNS_STD_ON_CHANGE_CONFIG)
    {
      AMS_DEV_PRINTF(MED, this, "AMS: ams_get_device_config: sensor_type %d (ON_CHANGE)",
                              sensor_type
                              );
      *chosen_sample_rate = 0;
      *chosen_report_rate = 0;
      *client_present = true;
      (*num_clients)++;
    }
    else if(request->message_id == SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_CONFIG)
    {
      AMS_DEV_PRINTF(MED, this, "AMS: ams_get_device_config: sensor_type %d (SELF TEST)",
                              sensor_type
                              );
      *chosen_sample_rate = state->odr_for_selftest;
      *chosen_report_rate = state->odr_for_selftest;
      *client_present = true;
    }
  }
}

static void ams_set_inst_config(sns_sensor *this,
                                    sns_sensor_instance *instance,
                                    float chosen_report_rate,
                                    float chosen_sample_rate,
                                    uint32_t message_id)
{
  sns_std_req new_client_config;
  sns_request config;

  new_client_config.desired_report_rate = chosen_report_rate;
  new_client_config.desired_sample_rate = chosen_sample_rate;

  config.message_id = message_id;
  config.request_len = sizeof(sns_std_req);
  config.request = &new_client_config;

  this->instance_api->set_client_config(instance, &config);
}
/*
 * For TOF devices, we don't turn power rail off to avoid losing RAM contents
 * we only put device to standby mode
 */
static void ams_turn_rails_off(sns_sensor *this)
{
  sns_sensor *sensor;

  for(sensor = this->cb->get_library_sensor(this, true);
      NULL != sensor;
      sensor = this->cb->get_library_sensor(this, false))
  {
    sns_time timeout = sns_convert_ns_to_ticks(AMS_POWER_RAIL_OFF_TIMEOUT_NS);
    tmf8801_start_power_rail_timer( sensor, timeout, AMS_POWER_RAIL_PENDING_OFF );
  }
  AMS_DEV_PRINTF(MED, this, "AMS: ams_turn_rails_off : prepare to turn off Rail");
}

void tmf8801_start_power_rail_timer(sns_sensor *const this,
                                sns_time timeout_ticks,
                                ams_power_rail_pending_state pwr_rail_pend_state)
{
  ams_state *state = (ams_state*)this->state->state;

  sns_timer_sensor_config req_payload = sns_timer_sensor_config_init_default;
  size_t req_len;
  uint8_t buffer[20];
  sns_memset(buffer, 0, sizeof(buffer));
  req_payload.is_periodic = false;
  req_payload.start_time = sns_get_system_time();
  req_payload.timeout_period = timeout_ticks;

  req_len = pb_encode_request(buffer, sizeof(buffer), &req_payload,
                              sns_timer_sensor_config_fields, NULL);
  if(req_len > 0)
  {
    sns_request timer_req =
      {  .message_id = SNS_TIMER_MSGID_SNS_TIMER_SENSOR_CONFIG,
         .request = buffer, .request_len = req_len  };
    state->timer_stream->api->send_request(state->timer_stream, &timer_req);
    state->power_rail_pend_state = pwr_rail_pend_state;
    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_start_power_rail_timer: enter... timer_stream %p",\
                        state->timer_stream);
  }
  else
  {
    AMS_PRINTF(HIGH, this, "AMS: tmf8801_start_power_rail_timer: timer req encode error");
  }
}

/**
 * Publish attributes read from registry
 *
 * @param[i] this    reference to this Sensor
 *
 * @return none
 */
static void
ams_publish_registry_attributes(sns_sensor *const this)
{
  int32_t hwid = this->cb->get_registration_index(this);
  ams_state *state = (ams_state*)this->state->state;
  {
    sns_std_attr_value_data value = sns_std_attr_value_data_init_default;
    value.has_boolean = true;
    value.boolean = state->supports_sync_stream;
    sns_publish_attribute(
        this, SNS_STD_SENSOR_ATTRID_STREAM_SYNC, &value, 1, false);
  }
  {
    sns_std_attr_value_data value = sns_std_attr_value_data_init_default;
    value.has_sint = true;
    value.sint = hwid;
    sns_publish_attribute(
        this, SNS_STD_SENSOR_ATTRID_HW_ID, &value, 1, false);
  }
  {
    sns_std_attr_value_data values[] = {SNS_ATTR, SNS_ATTR, SNS_ATTR, SNS_ATTR,
      SNS_ATTR, SNS_ATTR, SNS_ATTR, SNS_ATTR, SNS_ATTR, SNS_ATTR, SNS_ATTR, SNS_ATTR};
    for(uint8_t i = 0; i < 12; i++)
    {
      values[i].has_flt = true;
      values[i].flt = state->placement[i];
    }
    sns_publish_attribute(this, SNS_STD_SENSOR_ATTRID_PLACEMENT,
        values, ARR_SIZE(values), false);
  }
  {
    sns_std_attr_value_data value = sns_std_attr_value_data_init_default;
    value.has_sint = true;
    value.sint = state->registry_pf_cfg.rigid_body_type;
    sns_publish_attribute(
        this, SNS_STD_SENSOR_ATTRID_RIGID_BODY, &value, 1, true);
  }
}

void tmf8801_reval_instance_config(sns_sensor *this,
                              sns_sensor_instance *instance,
                              ams_sensor_type sensor_type)
{
  /**
   * Get best Config, decide best Instance Config based on above outputs.
   */
    UNUSED_VAR(sensor_type);
    float chosen_sample_rate = 0;
    float chosen_report_rate = 0;
    float sample_rate = 0;
    float report_rate = 0;
    bool client_present = false;
    int32_t num_clients;
    ams_state *state = (ams_state*)this->state->state;
    ams_instance_state *inst_state = (ams_instance_state*)instance->state->state;

    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_reval_instance_config: this %p, instance %p, sensor_type %d", this,
                          instance,
                          state->sensor_type);

    ams_get_device_config(this,
                           instance,
                           AMS_PROX_STRM,
                           &sample_rate,
                           &report_rate,
                           &client_present,
                           &num_clients);
    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_reval_instance_config: TOF: client_present %d, curr SR %d, recvd SR %d",
                            client_present,
                            (int32_t)chosen_sample_rate,
                            (int32_t)sample_rate);

    if(client_present)
    {
      chosen_sample_rate = SNS_MAX(chosen_sample_rate, sample_rate);
      chosen_report_rate = SNS_MAX(chosen_report_rate, report_rate);
      inst_state->publish_sensors |= AMS_PROX_STRM;
    }
    else
    {
      inst_state->publish_sensors &= ~(AMS_PROX_STRM);
    }

    ams_get_device_config(this,
                           instance,
                           AMS_PROX_OC,
                           &sample_rate,
                           &report_rate,
                           &client_present,
                           &num_clients);
    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_reval_instance_config: TOF: client_present %d, curr SR %d, recvd SR %d",
                            client_present,
                            (int32_t)chosen_sample_rate,
                            (int32_t)sample_rate);

    if (num_clients > inst_state->prox_info.client_cfg.num_oc_clients_prev
        &&
        num_clients > 1)
    {
      inst_state->prox_info.client_cfg.oc_send_last_event = true;
    }
    else
    {
      inst_state->prox_info.client_cfg.oc_send_last_event = false;
    }
    inst_state->prox_info.client_cfg.num_oc_clients_prev = num_clients;

    if(client_present)
    {
      chosen_sample_rate = SNS_MAX(chosen_sample_rate, sample_rate);
      chosen_report_rate = SNS_MAX(chosen_report_rate, report_rate);
      inst_state->publish_sensors |= AMS_PROX_OC;
    }
    else
    {
      inst_state->publish_sensors &= ~(AMS_PROX_OC);
    }

    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_reval_instance_config: TOF: publish_sensors=0x%04x",
                            inst_state->publish_sensors);

  if(state->sensor_type & AMS_STREAMERS)
  {
    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_reval_instance_config: sensor_type %d is a STREAMER, validate the ODR", state->sensor_type);
  }

  ams_set_inst_config(this,
                      instance,
                      chosen_report_rate,
                      chosen_sample_rate,
                      SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_CONFIG);

  if(!inst_state->publish_sensors)
  {
    AMS_DEV_PRINTF(MED, this, "AMS: tmf8801_reval_instance_config: .publish_sensors == 0, turn rails off, instance_is_ready_to_configure=FALSE");
    ams_turn_rails_off(this);
    inst_state->instance_is_ready_to_configure = false;
  }
}

/**
 * Sets instance config to run a self test.
 *
 * @param[i] this      Sensor reference
 * @param[i] instance  Sensor Instance reference
 *
 * @return none
 */
void tmf8801_set_self_test_inst_config(sns_sensor *this,
                              sns_sensor_instance *instance)
{

  sns_request config;

  config.message_id = SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_CONFIG;
  config.request_len = 0;
  config.request = NULL;

  this->instance_api->set_client_config(instance, &config);
}

/**
 * It extracts each element's version as well as it's value.
 *
 * @return bool True if decoding is successfull else false.
 */
bool
tmf8801_custom_parse_registry_float_array(sns_registry_data_item *reg_item,
                                      pb_buffer_arg  *item_name,
                                      pb_buffer_arg  *item_str_val,
                                      void *parsed_buffer)
{
  bool rv = true;
  pb_custom_float_parse_arg *arg = (pb_custom_float_parse_arg*) parsed_buffer;

  if(NULL == reg_item || NULL == item_name || NULL == item_str_val ||
     NULL == parsed_buffer)
  {
    rv = false;
  }
  else if(reg_item->has_flt && arg->size > *arg->index)
  {
    {
      arg->data_array[*arg->index] = reg_item->flt;
      if (reg_item->has_version) {
        arg->version_array[*arg->index] = reg_item->version;
      }
      (*arg->index)++;
    }
  }
  else
  {
    rv = false;
  }

  return rv;
}

bool
tmf8801_custom_parse_registry_string_array(sns_registry_data_item *reg_item,
                                      pb_buffer_arg  *item_name,
                                      pb_buffer_arg  *item_str_val,
                                      void *parsed_buffer)
{
  bool rv = true;
  pb_custom_string_parse_arg *arg = (pb_custom_string_parse_arg*) parsed_buffer;

  if(NULL == reg_item || NULL == item_name || NULL == item_str_val ||
     NULL == parsed_buffer)
  {
    rv = false;
  }
  else if(arg->size > *arg->index)
  {
      pb_buffer_arg *pb_buf = (pb_buffer_arg*)item_str_val;
      char *addr = (char*)pb_buf->buf;
      uint8_t lsize = STRING_PB_SIZE_MAX;
      if(pb_buf->buf_len < STRING_PB_SIZE_MAX)
        lsize = pb_buf->buf_len;
      for(int i= 0; i<lsize; i++)
      {
           global_parsed_char_array[i] = *addr++;
      }
      if (reg_item->has_version) {
        arg->version_array[*arg->index] = reg_item->version;
      }
      (*arg->index)++;
  }
  else
  {
    rv = false;
  }

  return rv;
}

bool
tmf8801_custom_parse_registry_ram_array(sns_registry_data_item *reg_item,
                                      pb_buffer_arg  *item_name,
                                      pb_buffer_arg  *item_str_val,
                                      void *parsed_buffer)
{
  bool rv = true;
  pb_custom_ram_fw_parse_arg *arg = (pb_custom_ram_fw_parse_arg*) parsed_buffer;
  if(NULL == reg_item || NULL == item_name || NULL == item_str_val ||
     NULL == parsed_buffer)
  {
    rv = false;
  }
  else
  {
      if(0 == strncmp((char*)item_name->buf, "record_count", item_name->buf_len)){

         ram_rec_size_total = (uint32_t)reg_item->sint;
      }
      else if(ram_rec_size_total > *arg->index)
      {
          pb_buffer_arg *pb_buf = (pb_buffer_arg*)item_str_val;
          char *addr = (char*)pb_buf->buf;
          uint8_t lsize = STRING_PB_SIZE_MAX;
          if(pb_buf->buf_len < STRING_PB_SIZE_MAX)
            lsize = pb_buf->buf_len;
          for(int i= 0; i<lsize; i++)
          {
               global_parsed_char_array[i] = *addr++;
          }
          (*arg->index)++;
          AMS_string_to_bytes(global_parsed_char_array, lsize, int_buf, INTEL_HEX_MAX_BYTES_PER_LINE);
          tmf8801_hexDecode_writeRam(arg->port_handle_ptr, int_buf);
      }
  }
  return rv;
}

bool
tmf8801_parse_registry_cmd_data_array(sns_registry_data_item *reg_item,
                                      pb_buffer_arg  *item_name,
                                      pb_buffer_arg  *item_str_val,
                                      void *parsed_buffer)
{
  bool rv = true;
  pb_tof_default_cmd_data_values *arg = (pb_tof_default_cmd_data_values*) parsed_buffer;

  if(NULL == reg_item || NULL == item_name || NULL == item_str_val ||
     NULL == parsed_buffer)
  {
    rv = false;
  }
  else if(reg_item->has_flt && arg->size > *arg->index)
  {
      arg->data_array[*arg->index] = reg_item->flt;
      if (reg_item->has_version) {
        arg->version_array[*arg->index] = reg_item->version;
      }

      pb_tof_default_cmd_data_values *has_values =
      (pb_tof_default_cmd_data_values *)parsed_buffer;

      if(0 == strncmp((char*)item_name->buf, "iteration", item_name->buf_len)){
        has_values->hasIteration = true;
      }
      else if(0 == strncmp((char*)item_name->buf, "period", item_name->buf_len)){
        has_values->hasPeriod = true;
      }
      else if(0 == strncmp((char*)item_name->buf, "threshold", item_name->buf_len)){
        has_values->hasThreshold = true;
      }
      else if(0 == strncmp((char*)item_name->buf, "delay", item_name->buf_len)){
        has_values->hasDelay = true;
      }
      else if(0 == strncmp((char*)item_name->buf, "gpio", item_name->buf_len)){
        has_values->hasGpio = true;
      }
      else if(0 == strncmp((char*)item_name->buf, "alg", item_name->buf_len)){
        has_values->hasAlg = true;
      }
      (*arg->index)++;
  }
  else
  {
    rv = false;
  }

  return rv;
}

static void ams_sensor_process_registry_event(sns_sensor *const this,
                                              sns_sensor_event *event)
{
  bool rv = true;
  ams_state *state = (ams_state*)this->state->state;
  float_t parsed_tof_cmd_data_array[TMF8801_CMD_DATA_REGISTRY_NUM] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

  pb_istream_t stream = pb_istream_from_buffer((void*)event->event,
      event->event_len);
  int32_t hwid = this->cb->get_registration_index(this);
  AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: enter: \
               this %p, state %p", this, state);
  if(SNS_REGISTRY_MSGID_SNS_REGISTRY_READ_EVENT == event->message_id)
  {
    sns_registry_read_event read_event = sns_registry_read_event_init_default;
    pb_buffer_arg group_name = {0,0};
    read_event.name.arg = &group_name;
    read_event.name.funcs.decode = pb_decode_string_cb;

    if(!pb_decode(&stream, sns_registry_read_event_fields, &read_event))
    {
      AMS_PRINTF(ERROR, this, "AMS: ams_sensor_process_registry_event: Error decoding registry event");
    }
    else
    {
      stream = pb_istream_from_buffer((void*)event->event, event->event_len);

      if(0 == strncmp((char*)group_name.buf, registry_tof_config_names[hwid], group_name.buf_len))
      {
        {
          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
              .group_name = "config",
              .parse_func = sns_registry_parse_phy_sensor_cfg,
              .parsed_buffer = &state->registry_cfg
            }
          };

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;

          rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
        }

        if(rv)
        {
          state->registry_cfg_received = true;
          state->is_dri = state->registry_cfg.is_dri;
          state->hardware_id = state->registry_cfg.hw_id;
          state->resolution_idx = state->registry_cfg.res_idx;
          state->supports_sync_stream = state->registry_cfg.sync_stream;
          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_tof_config received: \
               this %p, state %p", this, state);
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_process_registry_event: hardware_id:%d, is_dri: %d",
                     state->hardware_id, state->is_dri);
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_process_registry_event: resolution_idx:%d, supports_sync_stream:%d ",
                     state->resolution_idx, state->supports_sync_stream);
        }
      }
      else if(0 == strncmp((char*)group_name.buf, registry_platform_config_names[hwid],
                           group_name.buf_len))
      {
        {
          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
              .group_name = "config",
              .parse_func = sns_registry_parse_phy_sensor_pf_cfg,
              .parsed_buffer = &state->registry_pf_cfg
            }
          };

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;

          rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
        }

        if(rv)
        {
          state->registry_pf_cfg_received = true;

          state->com_port_info.com_config.bus_type = state->registry_pf_cfg.bus_type;
          state->com_port_info.com_config.bus_instance = state->registry_pf_cfg.bus_instance;
          state->com_port_info.com_config.slave_control = state->registry_pf_cfg.slave_config;
          state->com_port_info.com_config.min_bus_speed_KHz = state->registry_pf_cfg.min_bus_speed_khz;
          state->com_port_info.com_config.max_bus_speed_KHz = state->registry_pf_cfg.max_bus_speed_khz;
          state->com_port_info.com_config.reg_addr_type = state->registry_pf_cfg.reg_addr_type;
          state->irq_config.interrupt_num = state->registry_pf_cfg.dri_irq_num;
          state->irq_config.interrupt_pull_type = state->registry_pf_cfg.irq_pull_type;
          state->irq_config.is_chip_pin = state->registry_pf_cfg.irq_is_chip_pin;
          state->irq_config.interrupt_drive_strength = state->registry_pf_cfg.irq_drive_strength;
          state->irq_config.interrupt_trigger_type = state->registry_pf_cfg.irq_trigger_type;
          state->rail_config.num_of_rails = state->registry_pf_cfg.num_rail;
          state->registry_rail_on_state = state->registry_pf_cfg.rail_on_state;
          sns_strlcpy(state->rail_config.rails[0].name,
                      state->registry_pf_cfg.vddio_rail,
                      sizeof(state->rail_config.rails[0].name));
          sns_strlcpy(state->rail_config.rails[1].name,
                      state->registry_pf_cfg.vdd_rail,
                      sizeof(state->rail_config.rails[1].name));
          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_platform_config received: \
               this %p, state %p", this, state);
          AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: "
                     "bus_type:%d bus_instance:%d slave_control:%d",
                     state->com_port_info.com_config.bus_type,
                     state->com_port_info.com_config.bus_instance,
                     state->com_port_info.com_config.slave_control);

          AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: min_bus_speed_KHz :%d max_bus_speed_KHz:%d reg_addr_type:%d",
                               state->com_port_info.com_config.min_bus_speed_KHz,
                               state->com_port_info.com_config.max_bus_speed_KHz,
                               state->com_port_info.com_config.reg_addr_type);

          AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: interrupt_num:%d interrupt_pull_type:%d is_chip_pin:%d",
                               state->irq_config.interrupt_num,
                               state->irq_config.interrupt_pull_type,
                               state->irq_config.is_chip_pin);

          AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: interrupt_drive_strength:%d interrupt_trigger_type:%d"
             " rigid body type:%d",
             state->irq_config.interrupt_drive_strength,
             state->irq_config.interrupt_trigger_type,
             state->registry_pf_cfg.rigid_body_type);
        }
      }
      else if(0 == strncmp((char*)group_name.buf, registry_platform_placement_names[hwid],
                           group_name.buf_len))
      {
        {
          uint8_t arr_index = 0;
          pb_float_arr_arg arr_arg = {
            .arr = state->placement,
            .arr_index = &arr_index,
            .arr_len = 12
          };

          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
              .group_name = "placement",
              .parse_func = sns_registry_parse_float_arr,
              .parsed_buffer = &arr_arg
            }
          };

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;

          rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
        }

        if(rv)
        {
          state->registry_placement_received = true;
          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_platform_placement received: \
               this %p, state %p", this, state);
          AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: "
                 "p[0]:%d p[6]:%d p[11]:%d",
                 (uint32_t)state->placement[0], (uint32_t)state->placement[6],
                 (uint32_t)state->placement[11]);
        }
      }
      else if(0 == strncmp((char*)group_name.buf, registry_tof_fac_cal_names[hwid],
                           group_name.buf_len))
      {

        uint8_t version_array[1];
        {
          uint32_t arr_index = 0;

         pb_custom_string_parse_arg arr_arg = {
            .version_array = version_array,
            .index = &arr_index,
            .size = 1,
          };
          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
            .parse_func = tmf8801_custom_parse_registry_string_array,
            .parsed_buffer = &arr_arg
            }
          };
          strncpy(arg.parse_info[0].group_name, registry_tof_fac_cal_names[hwid], (SNS_REGISTRY_MAX_NAME_LEN - 1));

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;

          rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
        }
        if(rv)
        {
          state->sensorSystemData.calReceived = true;

          AMS_string_to_bytes(global_parsed_char_array, TMF8801_CALIBRATION_STRING_SIZE,
                            state->sensorSystemData.calBlock.calData, TMF8801_CALIBRATION_STRING_SIZE);
          state->cal_version[TOF_FAC_CAL_VERSION] = version_array[0];
          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_tof_fac_cal received: \
               this %p, state %p", this, state);
          AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: fac_cal: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x, 0x%x",
                   state->sensorSystemData.calBlock.calData[0], state->sensorSystemData.calBlock.calData[1],
                   state->sensorSystemData.calBlock.calData[2], state->sensorSystemData.calBlock.calData[3],
                   state->sensorSystemData.calBlock.calData[4], state->sensorSystemData.calBlock.calData[5],
                   state->sensorSystemData.calBlock.calData[6]);
          AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: fac_cal: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x, 0x%x",
                   state->sensorSystemData.calBlock.calData[7], state->sensorSystemData.calBlock.calData[8],
                   state->sensorSystemData.calBlock.calData[9], state->sensorSystemData.calBlock.calData[10],
                   state->sensorSystemData.calBlock.calData[11],state->sensorSystemData.calBlock.calData[12],
                   state->sensorSystemData.calBlock.calData[13]);
        }
      }
      else if(0 == strncmp((char*)group_name.buf, registry_tof_state_data_names[hwid],
                           group_name.buf_len))
      {
        uint8_t version_array[1];
        {
          uint32_t arr_index = 0;

         pb_custom_string_parse_arg arr_arg = {
            .version_array = version_array,
            .index = &arr_index,
            .size = 1,
          };
          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
              .parse_func = tmf8801_custom_parse_registry_string_array,
              .parsed_buffer = &arr_arg
            }
          };
          strncpy(arg.parse_info[0].group_name, registry_tof_state_data_names[hwid], (SNS_REGISTRY_MAX_NAME_LEN - 1));

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;

          rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
        }
        if(rv)
        {
          state->sensorSystemData.stateReceived = true;
          AMS_string_to_bytes(global_parsed_char_array, TMF8801_ALG_STATE_STRING_SIZE,
                            state->sensorSystemData.stateBlock.algStateData, TMF8801_ALG_STATE_DATA_SIZE);

          state->cal_version[TOF_STATE_DATA_VERSION] = version_array[0];

          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_tof_state_data received: \
               this %p, state %p", this, state);

          AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: state_data: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x",
                   state->sensorSystemData.stateBlock.algStateData[0], state->sensorSystemData.stateBlock.algStateData[1],
                   state->sensorSystemData.stateBlock.algStateData[2], state->sensorSystemData.stateBlock.algStateData[3],
                   state->sensorSystemData.stateBlock.algStateData[4],state->sensorSystemData.stateBlock.algStateData[5]);

          AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: state_data: 0x%x 0x%x 0x%x 0x%x 0x%x",
                   state->sensorSystemData.stateBlock.algStateData[6], state->sensorSystemData.stateBlock.algStateData[7],
                   state->sensorSystemData.stateBlock.algStateData[8],
                   state->sensorSystemData.stateBlock.algStateData[9], state->sensorSystemData.stateBlock.algStateData[10]);

        }
      }
      /**
       * THIS SECTION IS TO READ THE TOF DEFAULT CONFIG DATA FOR TOF FROM THE
       * JSON FILE
       */
      else if(0 == strncmp((char*)group_name.buf, registry_tof_cmd_data_names[hwid],
                           group_name.buf_len))
      {
        uint32_t version_array[TMF8801_CMD_DATA_REGISTRY_NUM] = {0, 0, 0, 0, 0, 0};
        struct pb_tof_default_cmd_data_values *pb_tof_cmd_data_value = NULL;
        {
          uint8_t arr_index = 0;

          struct pb_tof_default_cmd_data_values arr_arg = {
            .default_config_iteration   = 0.0,
            .default_config_period      = 0.0,
            .default_config_threshold   = 0.0,
            .default_config_delay       = 0.0,
            .default_config_gpio        = 0.0,
            .default_config_alg         = 0.0,
            .data_array = &parsed_tof_cmd_data_array[0],
            .version_array = version_array,
            .index = &arr_index,
            .size = TMF8801_CMD_DATA_REGISTRY_NUM,
            .hasIteration               = false,
            .hasPeriod                  = false,
            .hasThreshold               = false,
            .hasDelay                   = false,
            .hasGpio                    = false,
            .hasAlg                     = false
          };

          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
            .parse_func = tmf8801_parse_registry_cmd_data_array,
            .parsed_buffer = &arr_arg
            }
          };
          strncpy(arg.parse_info[0].group_name, registry_tof_cmd_data_names[hwid], (SNS_REGISTRY_MAX_NAME_LEN - 1));

          pb_tof_cmd_data_value = &arr_arg;

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;

          rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
        }
        if(rv)
        {
          /**
           * READ THE PROX DEFAULT CONFIG VALUES FROM THE JSON FILE
           */
          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_tof_cmd_data received: \
               this %p, state %p", this, state);
          if(true == pb_tof_cmd_data_value->hasIteration){
            uint16_t temp_short = (uint16_t)parsed_tof_cmd_data_array[0];
            state->sensorCmdData.iteration[0] = temp_short&0xFF;
            state->sensorCmdData.iteration[1] = (temp_short>>8)&0xFF;
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.cmd_data_iteration read from json %d", (uint32_t)parsed_tof_cmd_data_array[0]);
          }
          else{
            state->sensorCmdData.iteration[0] = default_tof_cmd_data.iteration[0];
            state->sensorCmdData.iteration[1] = default_tof_cmd_data.iteration[1];
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.cmd_data_iteration %d", (default_tof_cmd_data.iteration[1]<<8)|(default_tof_cmd_data.iteration[0]));
          }

          if(true == pb_tof_cmd_data_value->hasPeriod){
            state->sensorCmdData.period = (uint8_t)parsed_tof_cmd_data_array[1];
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.period read from json %d", (uint32_t)parsed_tof_cmd_data_array[1]);
          }
          else{
            state->sensorCmdData.period = default_tof_cmd_data.period;
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.period %d", default_tof_cmd_data.period);
          }

          if(true == pb_tof_cmd_data_value->hasThreshold){
            state->sensorCmdData.threshold = (uint8_t)parsed_tof_cmd_data_array[2];
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.threshold read from json %d", (uint32_t)parsed_tof_cmd_data_array[2]);
          }
          else{
            state->sensorCmdData.threshold = default_tof_cmd_data.threshold;
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.threshold %d", default_tof_cmd_data.threshold);
          }

          if(true == pb_tof_cmd_data_value->hasDelay){
            state->sensorCmdData.delay = (uint8_t)parsed_tof_cmd_data_array[3];
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.delay read from json %d", (uint32_t)parsed_tof_cmd_data_array[3]);
          }
          else{
            state->sensorCmdData.delay = default_tof_cmd_data.delay;
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.delay %d", default_tof_cmd_data.delay);
          }

          if(true == pb_tof_cmd_data_value->hasGpio){
            state->sensorCmdData.gpio = (uint8_t)parsed_tof_cmd_data_array[4];
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.gpio read from json %d", (uint32_t)parsed_tof_cmd_data_array[4]);
          }
          else{
            state->sensorCmdData.gpio = default_tof_cmd_data.gpio;
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.gpio %d", default_tof_cmd_data.gpio);
          }

          if(true == pb_tof_cmd_data_value->hasAlg){
            state->sensorCmdData.alg = (uint8_t)parsed_tof_cmd_data_array[5];
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.alg read from json %d", (int32_t)parsed_tof_cmd_data_array[5]);
          }
          else{
            state->sensorCmdData.alg = default_tof_cmd_data.alg;
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_process_registry_event: cmdData.alg %d", default_tof_cmd_data.alg);
          }
        }
      }
      else if(0 == strncmp((char*)group_name.buf, registry_tof_ram_names[hwid],
                           group_name.buf_len))
      {
        {
          uint32_t arr_index = 0;
          AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_process_registry_event: registry_tof_ram received: \
               this %p, state %p", this, state);
          pb_custom_ram_fw_parse_arg arr_arg = {
            .index = &arr_index,
            .port_handle_ptr = state->port_handle_ptr
          };
          sns_registry_decode_arg arg = {
            .item_group_name = &group_name,
            .parse_info_len = 1,
            .parse_info[0] = {
            .parse_func = tmf8801_custom_parse_registry_ram_array,
            .parsed_buffer = &arr_arg
            }
          };
          strncpy(arg.parse_info[0].group_name, registry_tof_ram_names[hwid], (SNS_REGISTRY_MAX_NAME_LEN - 1));

          read_event.data.items.funcs.decode = &sns_registry_item_decode_cb;
          read_event.data.items.arg = &arg;
          state->sensorSystemData.ramRegReceived = true;
          if(!ram_downloaded)
          {
              AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: Enable_Pon CPU_ready, prepare_for_RAM_downloading");

              tmf8801_devicePowerUp(state->port_handle_ptr, this);
#ifdef AMS_DEV_DEBUG_ENABLE
            {
              uint8_t TempByte;
              AMS_tmf8801_rwBlock(state->port_handle_ptr, TOF8801_REG_ENABLE, &TempByte, 1, I2C_READ);
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_process_registry_event: ctx->cpu_ready = 0x%x", TempByte);
              AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: start_RAM_downloading ...Rail is %d",
              (uint32_t)state->rail_config.rail_vote);
            }
#endif
              rv = pb_decode(&stream, sns_registry_read_event_fields, &read_event);
#ifdef AMS_DEV_DEBUG_ENABLE
           {
              uint8_t TempByte;
              AMS_DEV_PRINTF(LOW, this,"AMS: ams_sensor_process_registry_event: end_RAM_downloading ...Rail is %d",
              (uint32_t)state->rail_config.rail_vote);
              AMS_tmf8801_rwBlock(state->port_handle_ptr, TOF8801_REG_ENABLE, &TempByte, 1, I2C_READ);
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_process_registry_event: ctx->cpu_ready = 0x%x", TempByte);
           }
#endif
              tmf8801_download_systemData(state->port_handle_ptr, &state->sensorSystemData);
          }
          else
          {
            // receive the second ram patch registry, it is time to put device to standby_mode
            tmf8801_cpuGoStandby(state->port_handle_ptr);
          }

        }
        if(rv)
        {
          ram_downloaded = 1;
        }
      }
      else
      {
        rv = false;
      }

      if(!rv)
      {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_process_registry_event: error decoding registry group ");
      }
    }
  }
  else
  {
    AMS_PRINTF(ERROR, this, "AMS: ams_sensor_process_registry_event: Received unsupported registry event msg id %u",
                             event->message_id);
  }
}

void tmf8801_publish_available(sns_sensor *const this)
{
  sns_std_attr_value_data value = sns_std_attr_value_data_init_default;
  value.has_boolean = true;
  value.boolean = true;
  sns_publish_attribute(this, SNS_STD_SENSOR_ATTRID_AVAILABLE, &value, 1, true);
}

/** See sns_ams_sensor.h */
void tmf8801_process_suid_events(sns_sensor *const this)
{
  ams_state *state = (ams_state*)this->state->state;

  for(;
      0 != state->fw_stream->api->get_input_cnt(state->fw_stream);
      state->fw_stream->api->get_next_input(state->fw_stream))
  {
    sns_sensor_event *event =
      state->fw_stream->api->peek_input(state->fw_stream);

    if(SNS_SUID_MSGID_SNS_SUID_EVENT == event->message_id)
    {
      pb_istream_t stream = pb_istream_from_buffer((void*)event->event, event->event_len);
      sns_suid_event suid_event = sns_suid_event_init_default;
      pb_buffer_arg data_type_arg = { .buf = NULL, .buf_len = 0 };
      sns_sensor_uid uid_list;
      sns_suid_search suid_search;
      suid_search.suid = &uid_list;
      suid_search.num_of_suids = 0;

      suid_event.data_type.funcs.decode = &pb_decode_string_cb;
      suid_event.data_type.arg = &data_type_arg;
      suid_event.suid.funcs.decode = &pb_decode_suid_event;
      suid_event.suid.arg = &suid_search;

      if(!pb_decode(&stream, sns_suid_event_fields, &suid_event)) {
         AMS_PRINTF(ERROR, this, "AMS: ams_process_suid_events: SUID Decode failed");
         continue;
       }

      /* if no suids found, ignore the event */
      if(suid_search.num_of_suids == 0)
      {
        continue;
      }

      /* save suid based on incoming data type name */
      if(0 == strncmp(data_type_arg.buf, "interrupt", data_type_arg.buf_len))
      {
        state->irq_suid = uid_list;
      }
      else if(0 == strncmp(data_type_arg.buf, "timer", data_type_arg.buf_len))
      {
        state->timer_suid = uid_list;
      }
      else if (0 == strncmp(data_type_arg.buf, "registry", data_type_arg.buf_len))
      {
        state->reg_suid = uid_list;
      }
      else if(0 == strncmp(data_type_arg.buf, "async_com_port", data_type_arg.buf_len))
      {
        AMS_PRINTF(LOW, this, "AMS: ams_process_suid_events: received ASCP SUID");
      }
      else
      {
        AMS_PRINTF(ERROR, this, "AMS: ams_process_suid_events: invalid datatype_name");
      }
    }
  }
}

void tmf8801_sensor_send_registry_request(sns_sensor *const this,
                                             char *reg_group_name)
{
  ams_state *state = (ams_state*)this->state->state;
  uint8_t buffer[100];
  int32_t encoded_len;
  sns_memset(buffer, 0, sizeof(buffer));
  sns_rc rc = SNS_RC_SUCCESS;

  sns_registry_read_req read_request;
  pb_buffer_arg data = (pb_buffer_arg){ .buf = reg_group_name,
    .buf_len = (strlen(reg_group_name) + 1) };

  read_request.name.arg = &data;
  read_request.name.funcs.encode = pb_encode_string_cb;

  encoded_len = pb_encode_request(buffer, sizeof(buffer),
      &read_request, sns_registry_read_req_fields, NULL);
  if(0 < encoded_len)
  {
    sns_request request = (sns_request){
      .request_len = encoded_len, .request = buffer,
      .message_id = SNS_REGISTRY_MSGID_SNS_REGISTRY_READ_REQ };
    rc = state->reg_data_stream->api->send_request(state->reg_data_stream, &request);
  }
}

/** See sns_tmf8801_sensor.h*/
sns_rc tmf8801_sensor_notify_event(sns_sensor *const this)
{
  sns_sensor_event *event = NULL;

  ams_state *state = (ams_state*)this->state->state;
  sns_service_manager *service_mgr = this->cb->get_service_manager(this);
  sns_stream_service *stream_svc = (sns_stream_service*)
              service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);
  sns_time on_timestamp;
  sns_rc rv = SNS_RC_SUCCESS;
  sns_rc flag;
  ams_deviceCtx_t *amsLibrarytmf8801[tmf8801_NUM_DEVICES] = { NULL};
  int32_t hwid = this->cb->get_registration_index(this);

  AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_notify_event: enter: this %p, state %p, fw_stream %p, timer_stream %p, pwr_rail_service %p", this, state,\
                            state->fw_stream, state->timer_stream, state->pwr_rail_service);

  if(state->fw_stream)
  {
    if((0 == sns_memcmp(&state->irq_suid, &((sns_sensor_uid){{0}}), sizeof(state->irq_suid)))
     || (0 == sns_memcmp(&state->timer_suid, &((sns_sensor_uid){{0}}), sizeof(state->timer_suid)))
     || (0 == sns_memcmp(&state->reg_suid, &((sns_sensor_uid){{0}}), sizeof(state->reg_suid)))
    )
    {
      tmf8801_process_suid_events(this);

      // place a request to registry sensor
      if((0 != sns_memcmp(&state->reg_suid, &((sns_sensor_uid){{0}}), sizeof(state->reg_suid))))
      {
        if(state->reg_data_stream == NULL)
        {
          stream_svc->api->create_sensor_stream(stream_svc,
              this, state->reg_suid , &state->reg_data_stream);

          tmf8801_sensor_send_registry_request(this, registry_platform_config_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_platform_placement_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_config_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_fac_cal_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_cmd_data_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_state_data_names[hwid]);
        }
      }
    }
  }

  /**----------------------Handle a Timer Sensor event.-------------------*/
  if(NULL != state->timer_stream)
  {
    AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_notify_event: timer_stream");

    event = state->timer_stream->api->peek_input(state->timer_stream);

    while(NULL != event)
    {
      pb_istream_t stream = pb_istream_from_buffer((pb_byte_t*)event->event,
                                                   event->event_len);
      sns_timer_sensor_event timer_event;
      if(event->message_id == SNS_TIMER_MSGID_SNS_TIMER_SENSOR_EVENT)
      {
        if(pb_decode(&stream, sns_timer_sensor_event_fields, &timer_event))
        {
          if(state->power_rail_pend_state == AMS_POWER_RAIL_PENDING_INIT)
          {
            /**-------------------Read and Confirm WHO-AM-I------------------------*/
            AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_notify_event: try amsLibrary_open_tmf8801...");

            state->priv = amsLibrary_open_tmf8801(state->port_handle_ptr,
                                                  &state->sensorSystemData,
                                                  &state->sensorCmdData,
                                                  amsLibrarytmf8801[hwid],
                                                  hwid);

            if (state->priv != NULL){
              AMS_DEV_PRINTF(LOW, this, "AMS: ams_sensor_notify_event: amsLibrary_open_tmf8801 SUCCESS");
              state->tmf8801_hw_is_present = true;
              // request ram patch here to guarantee rail is on and hardware available
              tmf8801_sensor_send_registry_request(this, registry_tof_ram_names[hwid]);
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: state->tmf8801_hw_is_present %d", state->tmf8801_hw_is_present);
            } else {
              AMS_PRINTF(ERROR, this, "AMS: ams_sensor_notify_event: amsLibrary_open_tmf8801 ERROR");
              rv = SNS_RC_INVALID_LIBRARY_STATE;
            }
            if(state->tmf8801_hw_is_present)
            {
                AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_notify_event: HW present ");
                tmf8801_publish_available(this);
            }
            else
            {
              rv = SNS_RC_INVALID_LIBRARY_STATE;
              AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_notify_event: HW absent");
            }

            state->power_rail_pend_state = AMS_POWER_RAIL_PENDING_NONE;
          }
          else if(state->power_rail_pend_state == AMS_POWER_RAIL_PENDING_SET_CLIENT_REQ)
          {
            AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_notify_event: _POWER_RAIL_PENDING_SET_CLIENT_REQ");

            sns_sensor_instance *instance = sns_sensor_util_get_shared_instance(this);
            AMS_DEV_PRINTF(MED,this, "AMS: ams_sensor_notify_event: shared instance %p", instance);
            if(NULL != instance)
            {
              ams_instance_state *inst_state =
                 (ams_instance_state*) instance->state->state;
              inst_state->instance_is_ready_to_configure = true;
              AMS_DEV_PRINTF(MED,this, "AMS: ams_sensor_notify_event: instance_is_ready_to_configure=%d"
                                , inst_state->instance_is_ready_to_configure);
              tmf8801_reval_instance_config(this, instance, state->sensor_type);
              if(inst_state->new_self_test_request)
              {
                tmf8801_set_self_test_inst_config(this, instance);
              }
            }
            state->power_rail_pend_state = AMS_POWER_RAIL_PENDING_NONE;
          }
          else if(AMS_POWER_RAIL_PENDING_OFF == state->power_rail_pend_state)
          {
            sns_sensor_instance *instance = sns_sensor_util_get_shared_instance(this);
            /*Not required to turn OFF power rail, If instance is present*/
            if(NULL == instance)
            {
                //tmf8801_cpuGoStandby(state->port_handle_ptr);
                //AMS_DEV_PRINTF(MED, this, "AMS: ams_sensor_notify_event : put TOF device to stand_by mode");
            }
            state->power_rail_pend_state = AMS_POWER_RAIL_PENDING_NONE;
          }
        }
        else
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: pb_decode error");
        }
      }
      event = state->timer_stream->api->get_next_input(state->timer_stream);
    }
  }

  if(NULL != state->reg_data_stream)
  {
    event = state->reg_data_stream->api->peek_input(state->reg_data_stream);
    while(NULL != event)
    {
      ams_sensor_process_registry_event(this, event);
      if(state->registry_cfg_received && state->registry_placement_received)
      {
        ams_publish_registry_attributes(this);
      }
      event = state->reg_data_stream->api->get_next_input(state->reg_data_stream);
    }
  }

  /**---------------------Register Power Rails --------------------------*/
  if(0 != sns_memcmp(&state->timer_suid, &((sns_sensor_uid){{0}}), sizeof(state->timer_suid))
     && state->registry_pf_cfg_received
     && state->registry_cfg_received
     && state->registry_placement_received)
  {
    state->registry_pf_cfg_received = false;

    /**-----------------Register and Open COM Port-------------------------*/
    if(NULL == state->com_port_info.port_handle)
    {
      AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: .port_handle is null");
      rv = state->scp_service->api->sns_scp_register_com_port(&state->com_port_info.com_config,
                                                        &state->com_port_info.port_handle);
      AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: .port_handle %p, set AMS_PORT_portHndl", state->com_port_info.port_handle);

      if(SNS_RC_SUCCESS == rv)
      {
        rv = state->scp_service->api->sns_scp_open(state->com_port_info.port_handle);
        if(SNS_RC_SUCCESS == rv)
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: sns_scp_open SUCCESS");
          /* set our wrapped-up handle for this port */
          state->port_handle.scp_handle = state->com_port_info.port_handle;
          state->port_handle_ptr = &state->port_handle;
        }
        else
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: sns_scp_open failed");
        }
      }
    }
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: register power rails");
    state->rail_config.rail_vote = SNS_RAIL_OFF;

    state->pwr_rail_service =
     (sns_pwr_rail_service*)service_mgr->get_service(service_mgr,
                                                     SNS_POWER_RAIL_SERVICE);

    state->pwr_rail_service->api->sns_register_power_rails(state->pwr_rail_service,
                                                           &state->rail_config);

    /**---------------------Turn Power Rails ON----------------------------*/
    state->rail_config.rail_vote = SNS_RAIL_ON_NPM;
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: turn Rail On ");
    flag = state->pwr_rail_service->api->sns_vote_power_rail_update(state->pwr_rail_service,
                                                             this,
                                                             &state->rail_config,
                                                             &on_timestamp);
    if(flag == SNS_RC_SUCCESS){
        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: Turn on Rail");
    }
    else{
         AMS_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: rail turn on failed");
    }
    /**-------------Create a Timer stream for Power Rail ON timeout.---------*/
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: state->tmf8801_hw_is_present %d", state->tmf8801_hw_is_present);
    if(NULL == state->timer_stream)
    {
      AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: .timer_stream is null");
      stream_svc->api->create_sensor_stream(stream_svc, this, state->timer_suid,
                                            &state->timer_stream);
      AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: .timer_stream %p", state->timer_stream);
      if(NULL != state->timer_stream)
     {
        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_sensor_notify_event: power rail timer start");

        tmf8801_start_power_rail_timer(this,
                                  sns_convert_ns_to_ticks(AMS_DEVICE_OFF_TO_IDLE_MS * 1000 * 1000),
                                  AMS_POWER_RAIL_PENDING_INIT);
      }
    }
  }
  return rv;
}

sns_rc tmf8801_streaming_sensor_notify_event(sns_sensor *const this)
{
  ams_state *state = (ams_state*)this->state->state;
  sns_service_manager *service_mgr = this->cb->get_service_manager(this);
  sns_stream_service *stream_svc = (sns_stream_service*)
              service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);
  sns_time on_timestamp;
  sns_rc rv = SNS_RC_SUCCESS;
  sns_rc flag;
  sns_sensor_event *event;
  ams_deviceCtx_t *amsLibrarytmf8801[tmf8801_NUM_DEVICES] = { NULL};
  int32_t hwid = this->cb->get_registration_index(this);

  AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: enter: this %p, state %p, fw_stream %p, timer_stream %p, pwr_rail_service %p",
             this, state, state->fw_stream, state->timer_stream, state->pwr_rail_service);

  if(state->fw_stream)
  {
    if((0 == sns_memcmp(&state->irq_suid, &((sns_sensor_uid){{0}}), sizeof(state->irq_suid)))
     || (0 == sns_memcmp(&state->timer_suid, &((sns_sensor_uid){{0}}), sizeof(state->timer_suid)))
     || (0 == sns_memcmp(&state->reg_suid, &((sns_sensor_uid){{0}}), sizeof(state->reg_suid)))
    )
    {
      tmf8801_process_suid_events(this);
      // place a request to registry sensor
      if((0 != sns_memcmp(&state->reg_suid, &((sns_sensor_uid){{0}}), sizeof(state->reg_suid))))
      {
        if(state->reg_data_stream == NULL)
        {
          stream_svc->api->create_sensor_stream(stream_svc,
              this, state->reg_suid , &state->reg_data_stream);

          tmf8801_sensor_send_registry_request(this, registry_platform_config_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_platform_placement_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_config_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_fac_cal_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_cmd_data_names[hwid]);
          tmf8801_sensor_send_registry_request(this, registry_tof_state_data_names[hwid]);
        }
      }
    }
  }

  /**----------------------Handle a Timer Sensor event.-------------------*/
  if(NULL != state->timer_stream)
  {
    AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: timer_stream");

    event = state->timer_stream->api->peek_input(state->timer_stream);

    while(NULL != event)
    {
      pb_istream_t stream = pb_istream_from_buffer((pb_byte_t*)event->event,
                                                   event->event_len);
      sns_timer_sensor_event timer_event;
      if(event->message_id == SNS_TIMER_MSGID_SNS_TIMER_SENSOR_EVENT)
      {
        if(pb_decode(&stream, sns_timer_sensor_event_fields, &timer_event))
        {
          if(state->power_rail_pend_state == AMS_POWER_RAIL_PENDING_INIT)
          {
            /**-------------------Read and Confirm WHO-AM-I------------------------*/
            AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: try amsLibrary_open_tmf8801...");

            state->priv = amsLibrary_open_tmf8801(state->port_handle_ptr,
                                                  &state->sensorSystemData,
                                                  &state->sensorCmdData,
                                                  amsLibrarytmf8801[hwid],
                                                  hwid);

            if (state->priv != NULL){
              AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: amsLibrary_open_tmf8801 SUCCESS");
              state->tmf8801_hw_is_present = true;
              // request ram patch here to guarantee rail is on and hardware available
              tmf8801_sensor_send_registry_request(this, registry_tof_ram_names[hwid]);
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: state->tmf8801_hw_is_present %d", state->tmf8801_hw_is_present);
            } else {
              AMS_PRINTF(ERROR, this, "AMS: ams_streaming_sensor_notify_event: amsLibrary_open_tmf8801 ERROR");
              rv = SNS_RC_INVALID_LIBRARY_STATE;
            }

            if(state->tmf8801_hw_is_present)
            {
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: HW present");
              tmf8801_publish_available(this);
            }
            else
            {
              rv = SNS_RC_INVALID_LIBRARY_STATE;
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: HW absent");
            }
            state->power_rail_pend_state = AMS_POWER_RAIL_PENDING_NONE;
          }
          else if(state->power_rail_pend_state == AMS_POWER_RAIL_PENDING_SET_CLIENT_REQ)
          {
            AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: _POWER_RAIL_PENDING_SET_CLIENT_REQ");

            sns_sensor_instance *instance = sns_sensor_util_get_shared_instance(this);
            AMS_DEV_PRINTF(MED,this, "AMS: ams_streaming_sensor_notify_event: shared instance %p", instance);
            if(NULL != instance)
            {
              ams_instance_state *inst_state =
                 (ams_instance_state*) instance->state->state;
              inst_state->instance_is_ready_to_configure = true;
              AMS_DEV_PRINTF(MED,this, "AMS: ams_streaming_sensor_notify_event: instance_is_ready_to_configure=%d"
                                , inst_state->instance_is_ready_to_configure);
              tmf8801_reval_instance_config(this, instance, state->sensor_type);
              if(inst_state->new_self_test_request)
              {
                tmf8801_set_self_test_inst_config(this, instance);
              }
            }
            state->power_rail_pend_state = AMS_POWER_RAIL_PENDING_NONE;
          }
          else if(AMS_POWER_RAIL_PENDING_OFF == state->power_rail_pend_state)
          {
            sns_sensor_instance *instance = sns_sensor_util_get_shared_instance(this);
            /*Not required to turn OFF power rail, If instance is present*/
            if(NULL == instance)
            {
                //tmf8801_cpuGoStandby(state->port_handle_ptr);
                //AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event : put TOF device to stand_by mode");
            }
            state->power_rail_pend_state = AMS_POWER_RAIL_PENDING_NONE;
          }
        }
        else
        {
          AMS_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: pb_decode error");
        }
      }
      event = state->timer_stream->api->get_next_input(state->timer_stream);
    }
  }
  if(NULL != state->reg_data_stream)
  {
    event = state->reg_data_stream->api->peek_input(state->reg_data_stream);
    while(NULL != event)
    {
      ams_sensor_process_registry_event(this, event);
      if(state->registry_cfg_received && state->registry_placement_received)
      {
        ams_publish_registry_attributes(this);
      }
      event = state->reg_data_stream->api->get_next_input(state->reg_data_stream);
    }
  }

  /**---------------------Register Power Rails --------------------------*/
  if(0 != sns_memcmp(&state->timer_suid, &((sns_sensor_uid){{0}}), sizeof(state->timer_suid))
     && state->registry_pf_cfg_received
     && state->registry_cfg_received
     && state->registry_placement_received)
  {
    state->registry_pf_cfg_received = false;

    /**-----------------Register and Open COM Port-------------------------*/
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: state->tmf8801_hw_is_present %d", state->tmf8801_hw_is_present);
    if((NULL == state->com_port_info.port_handle)/*  && (state->tmf8801_hw_is_present)*/)
    {
      AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: .port_handle is null");
      rv = state->scp_service->api->sns_scp_register_com_port(&state->com_port_info.com_config,
                                                        &state->com_port_info.port_handle);
      AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: .port_handle %p, set AMS_PORT_portHndl", state->com_port_info.port_handle);

      if(rv == SNS_RC_SUCCESS)
      {
        rv = state->scp_service->api->sns_scp_open(state->com_port_info.port_handle);
        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: sns_scp_open SUCCESS");
        /* set our wrapped-up handle for this port */
        state->port_handle.scp_handle = state->com_port_info.port_handle;
        state->port_handle_ptr = &state->port_handle;
      }
      else
      {
        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: sns_scp_open failed");
      }
    }
    AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: register power rails");
    state->rail_config.rail_vote = SNS_RAIL_OFF;

    state->pwr_rail_service =
     (sns_pwr_rail_service*)service_mgr->get_service(service_mgr,
                                                     SNS_POWER_RAIL_SERVICE);

    state->pwr_rail_service->api->sns_register_power_rails(state->pwr_rail_service,
                                                           &state->rail_config);

    /**---------------------Turn Power Rails ON----------------------------*/
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: state->tmf8801_hw_is_present %d", state->tmf8801_hw_is_present);
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: registry cfg sections rcvd:  vote for SNS_RAIL_ON_NPM");
    state->rail_config.rail_vote = SNS_RAIL_ON_NPM;
    AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: Turn on Rail");
    flag = state->pwr_rail_service->api->sns_vote_power_rail_update(state->pwr_rail_service,
                                                             this,
                                                             &state->rail_config,
                                                             &on_timestamp);

    if(flag == SNS_RC_SUCCESS){
        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: Turn on Rail");
    }
    else{
        AMS_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: rail turn on failed");
    }

    /**-------------Create a Timer stream for Power Rail ON timeout.---------*/
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_streaming_sensor_notify_event: state->tmf8801_hw_is_present %d", state->tmf8801_hw_is_present);
    if(NULL == state->timer_stream)
    {
      AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: .timer_stream is null");
      stream_svc->api->create_sensor_stream(stream_svc, this, state->timer_suid,
                                            &state->timer_stream);
      AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: .timer_stream %p", state->timer_stream);
      if(NULL != state->timer_stream)
      {
        AMS_DEV_PRINTF(MED, this, "AMS: ams_streaming_sensor_notify_event: power rail timer start");
        tmf8801_start_power_rail_timer(this,
                                  sns_convert_ns_to_ticks(AMS_DEVICE_OFF_TO_IDLE_MS * 1000 * 1000),
                                  AMS_POWER_RAIL_PENDING_INIT);
      }
    }
  }
  return rv;
}

static bool
tmf8801_write_to_tof_fac_cal(struct pb_ostream_s *stream,
    struct pb_field_s const *field, void *const *arg)
{
    sns_sensor *this = (sns_sensor*) *arg;

    ams_state *state = (ams_state *)this->state->state;
    tof_calibration_data_t *cfg = &state->sensorSystemData.calBlock;
    uint32_t *new_version = state->cal_version;

    char const *names[] = {"calibration"};
    char  *data[] = {"0102030405060708090A0B0C0D0E"};
    char  str[TMF8801_ALG_STATE_STRING_SIZE];

    AMS_bytes_to_string(cfg->calData, TMF8801_CALIBRATION_DATA_SIZE, str, TMF8801_CALIBRATION_STRING_SIZE);
    strncpy(data[0], str, TMF8801_CALIBRATION_STRING_SIZE);

    pb_buffer_arg name_data = (pb_buffer_arg)
           { .buf = names[0], .buf_len = strlen(names[0]) + 1 };

    pb_buffer_arg str_data = (pb_buffer_arg)
           { .buf = data[0], .buf_len = strlen(data[0]) + 1};
    /* QC - consider using sizeof() instead of strlen() + 1. Comment applies to all such usage through driver.*/
    sns_registry_data_item pb_item = sns_registry_data_item_init_default;

    pb_item.name.funcs.encode = &pb_encode_string_cb;
    pb_item.name.arg = &name_data;

    pb_item.has_version = true;
    pb_item.version = new_version[TOF_FAC_CAL_VERSION];
    pb_item.str.funcs.encode = &pb_encode_string_cb;
    pb_item.str.arg = &str_data;

    AMS_DEV_PRINTF(LOW, this,"AMS: tmf8801_write_to_tof_fac_cal: %d %d %d %d %d %d %d ",
                   cfg->calData[0], cfg->calData[1], cfg->calData[2], cfg->calData[3],
                   cfg->calData[4], cfg->calData[5], cfg->calData[6]);

    AMS_DEV_PRINTF(LOW, this,"AMS: tmf8801_write_to_tof_fac_cal: %d %d %d %d %d %d %d ",
                   cfg->calData[7],cfg->calData[8], cfg->calData[9], cfg->calData[10],
                   cfg->calData[11],cfg->calData[12], cfg->calData[13]);

    if(!pb_encode_tag_for_field(stream, field))
    {
      AMS_PRINTF(ERROR, this,"AMS: tmf8801_write_to_tof_fac_cal: Error encoding tag item");
      return false;
    }

    if(!pb_encode_submessage(stream, sns_registry_data_item_fields, &pb_item))
    {
      AMS_PRINTF(ERROR, this,"AMS: tmf8801_write_to_tof_fac_cal: Error encoding item");
      return false;
    }

    AMS_DEV_PRINTF(HIGH, this, "AMS: tmf8801_write_to_tof_fac_cal: done");
    return true;
}

static bool
tmf8801_write_to_tof_state_data(struct pb_ostream_s *stream,
    struct pb_field_s const *field, void *const *arg)
{
    sns_sensor *this = (sns_sensor*) *arg;

    ams_state *state = (ams_state *)this->state->state;
    tof_alg_state_data_t *cfg = &state->sensorSystemData.stateBlock;
    uint32_t *new_version = state->cal_version;

    char const *names[] = {"state_data"};
    char  *data[] = {"0102030405060708090A0B"};
    char  str[TMF8801_ALG_STATE_STRING_SIZE];

    AMS_bytes_to_string(cfg->algStateData, TMF8801_ALG_STATE_DATA_SIZE, str, TMF8801_ALG_STATE_STRING_SIZE);
    strncpy(data[0], str, TMF8801_ALG_STATE_STRING_SIZE);

    pb_buffer_arg name_data = (pb_buffer_arg)
           { .buf = names[0], .buf_len = strlen(names[0]) + 1 };

    pb_buffer_arg str_data = (pb_buffer_arg)
           { .buf = data[0], .buf_len = strlen(data[0]) + 1};
    /* QC - consider using sizeof() instead of strlen() + 1. Comment applies to all such usage through driver.*/
    sns_registry_data_item pb_item = sns_registry_data_item_init_default;

    pb_item.name.funcs.encode = &pb_encode_string_cb;
    pb_item.name.arg = &name_data;

    pb_item.has_version = true;
    pb_item.version = new_version[TOF_STATE_DATA_VERSION];
    pb_item.str.funcs.encode = &pb_encode_string_cb;
    pb_item.str.arg = &str_data;

    AMS_DEV_PRINTF(HIGH, this,"AMS: tmf8801_write_to_tof_state_data: %d %d %d %d %d",
                   cfg->algStateData[0], cfg->algStateData[1], cfg->algStateData[2],
                   cfg->algStateData[3], cfg->algStateData[4]);
    AMS_DEV_PRINTF(HIGH, this,"AMS: tmf8801_write_to_tof_state_data: %d %d %d %d %d %d",
                   cfg->algStateData[5], cfg->algStateData[6], cfg->algStateData[7],
                   cfg->algStateData[8], cfg->algStateData[9], cfg->algStateData[10]);
    if(!pb_encode_tag_for_field(stream, field))
    {
      AMS_PRINTF(ERROR, this,"AMS: tmf8801_write_to_tof_state_data: Error encoding tag item:");
      return false;
    }

    if(!pb_encode_submessage(stream, sns_registry_data_item_fields, &pb_item))
    {
      AMS_PRINTF(ERROR, this,"AMS: tmf8801_write_to_tof_state_data: Error encoding item ");
      return false;
    }
    AMS_DEV_PRINTF(HIGH, this, "AMS: tmf8801_write_to_tof_state_data: done");
    return true;
}

static bool
tmf8801_write_calibration_to_registry(sns_sensor *const this)
{
  //todo how much to define for this buffer
  uint8_t buffer[1000];
  int32_t encoded_len;
  ams_state *state =  (ams_state *) this->state->state;
  sns_registry_write_req write_req = sns_registry_write_req_init_default;
  int32_t hwid = this->cb->get_registration_index(this);

  /* Should never happen, but just in case */
  if ((hwid > (tmf8801_NUM_DEVICES - 1)) || (hwid < 0))
      return false;

  const char *names[] = { AMS_REGISTRY_ENTRY_NAME("_platform.tof.fac_cal"),
  AMS_REGISTRY_ENTRY_NAME("_platform.tof.state_data")};


  for (int i = 0 ; i < sizeof(names)/sizeof(names[0]) ; i++)
  {
    pb_buffer_arg name_data = (pb_buffer_arg)
          { .buf = names[i],
            .buf_len = strlen (names[i]) + 1 };

    write_req.name.funcs.encode = &pb_encode_string_cb;
    write_req.name.arg = &name_data;
    write_req.data.items.arg = this;

    if (0 == strncmp(names[i],
          AMS_REGISTRY_ENTRY_NAME("_platform.tof.fac_cal"),
          sizeof(AMS_REGISTRY_ENTRY_NAME("_platform.tof.fac_cal"))))
    {
      write_req.data.items.funcs.encode = tmf8801_write_to_tof_fac_cal;
    }
    else if (0 == strncmp(names[i],
          AMS_REGISTRY_ENTRY_NAME("_platform.tof.state_data"),
          sizeof(AMS_REGISTRY_ENTRY_NAME("_platform.tof.state_data"))))
    {
      write_req.data.items.funcs.encode = tmf8801_write_to_tof_state_data;
    }

    encoded_len = pb_encode_request(buffer, sizeof(buffer),
        &write_req, sns_registry_write_req_fields, NULL);
    if(0 < encoded_len)
    {
      sns_request request = (sns_request){
        .request_len = encoded_len, .request = buffer,
            .message_id = SNS_REGISTRY_MSGID_SNS_REGISTRY_WRITE_REQ };
      state->reg_data_stream->api->send_request(state->reg_data_stream, &request);
    } else {
      return false;
    }
  }

  return true;
}

/**
 * Decodes self test requests.
 *
 * @param[i] this              Sensor reference
 * @param[i] request           Encoded input request
 * @param[o] decoded_request   Decoded standard request
 * @param[o] test_config       decoded self test request
 *
 * @return bool True if decoding is successfull else false.
 */

static bool ams_get_decoded_self_test_request(sns_sensor const *this, sns_request const *request,
                                                  sns_std_request *decoded_request,
                                                  sns_physical_sensor_test_config *test_config)
{
  pb_istream_t stream;
  pb_simple_cb_arg arg =
      { .decoded_struct = test_config,
        .fields = sns_physical_sensor_test_config_fields };
  decoded_request->payload = (struct pb_callback_s)
      { .funcs.decode = &pb_decode_simple_cb, .arg = &arg };
  stream = pb_istream_from_buffer(request->request,
                                  request->request_len);
  if(!pb_decode(&stream, sns_std_request_fields, decoded_request))
  {
    AMS_PRINTF(ERROR, this, "AMS: ams_get_decoded_self_test_request: pb_decode error");
    return false;
  }
  return true;
}

/**
 * Decodes a physical sensor self test request and updates
 * instance state with request info.
 *
 * @param[i] this      Sensor reference
 * @param[i] instance  Sensor Instance reference
 * @param[i] new_request Encoded request
 *
 * @return True if request is valid else false
 */
static bool ams_extract_self_test_info(sns_sensor *this,
                              sns_sensor_instance *instance,
                              struct sns_request const *new_request)
{
  sns_std_request decoded_request = sns_std_request_init_default;
  sns_physical_sensor_test_config test_config = sns_physical_sensor_test_config_init_default;
  ams_state *state = (ams_state*)this->state->state;
  ams_instance_state *inst_state = (ams_instance_state*)instance->state->state;
  ams_self_test_info *self_test_info;

  if (state->sensor_type == AMS_PROX_OC) {
    self_test_info = &inst_state->prox_info.test_info[AMS_OC_IDX];
  }
  else if (state->sensor_type == AMS_PROX_STRM) {
    self_test_info = &inst_state->prox_info.test_info[AMS_STRM_IDX];
  }
  else
  {
    AMS_PRINTF(ERROR, this, "AMS: ams_extract_self_test_info: Error: unknown sensor type");
    return false;
  }

  if(ams_get_decoded_self_test_request(this, new_request, &decoded_request, &test_config))
  {
    self_test_info->test_type = test_config.test_type;
    self_test_info->test_client_present = true;
    return true;
  }
  else
  {
    return false;
  }
}

void tmf8801_send_suid_req(sns_sensor *this, char *const data_type,
    uint32_t data_type_len)
{
  uint8_t buffer[50];
  sns_memset(buffer, 0, sizeof(buffer));
  ams_state *state = (ams_state*)this->state->state;
  sns_service_manager *manager = this->cb->get_service_manager(this);
  sns_stream_service *stream_service =
    (sns_stream_service*)manager->get_service(manager, SNS_STREAM_SERVICE);
  size_t encoded_len;
  pb_buffer_arg data = (pb_buffer_arg){ .buf = data_type, .buf_len = data_type_len };

  sns_suid_req suid_req = sns_suid_req_init_default;
  suid_req.has_register_updates = true;
  suid_req.register_updates = true;
  suid_req.data_type.funcs.encode = &pb_encode_string_cb;
  suid_req.data_type.arg = &data;

  if(state->fw_stream == NULL)
  {
     stream_service->api->create_sensor_stream(stream_service,
         this, sns_get_suid_lookup(), &state->fw_stream);
  }

  encoded_len = pb_encode_request(buffer, sizeof(buffer),
      &suid_req, sns_suid_req_fields, NULL);
  if(0 < encoded_len)
  {
    sns_request request = (sns_request){
      .request_len = encoded_len, .request = buffer, .message_id = SNS_SUID_MSGID_SNS_SUID_REQ };
    state->fw_stream->api->send_request(state->fw_stream, &request);
  }
}

/** See sns_ams_sensor.h */
sns_sensor_instance* tmf8801_set_client_request(sns_sensor *const this,
                                                struct sns_request const *exist_request,
                                                struct sns_request const *new_request,
                                                bool remove)
{
  sns_sensor_instance *instance = sns_sensor_util_get_shared_instance(this);
  ams_state *state = (ams_state*)this->state->state;
  sns_time on_timestamp;
  sns_time delta;
  bool reval_config = false;
  int32_t hwid = this->cb->get_registration_index(this);
  sns_rc flag;

#ifdef ENABLE_DEBUG
  if (remove)
  {
      AMS_DEV_PRINTF(HIGH, this
                 , "AMS: ams_set_client_request: enter: REMOVE: sensor_this 0x%X, exist_rq 0x%X, new_rq 0x%X; inst 0x%X"
                 , this, exist_request, new_request, instance);
  } else {
      AMS_DEV_PRINTF(HIGH, this
                 , "AMS: ams_set_client_request: enter: ADD/CHANGE: sensor_this 0x%X, exist_rq 0x%X, new_rq 0x%X; inst 0x%X"
                 , this, exist_request, new_request, instance);
  }
#endif


  sns_sensor_uid * p_prox_oc_suid = (sns_sensor_uid *)tmf8801_get_suid_with_sensortype_and_hwid(AMS_PROX_OC, hwid);
  sns_sensor_uid * p_prox_st_suid = (sns_sensor_uid *)tmf8801_get_suid_with_sensortype_and_hwid(AMS_PROX_STRM, hwid);


  if(remove)
  {
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request:  remove rqst; instance 0x%X", instance);
    if(NULL != instance)
    {
      ams_instance_state *inst_state = (ams_instance_state *) instance->state->state;
      instance->cb->remove_client_request(instance, exist_request);

      if (exist_request->message_id !=
          SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_CONFIG)
      {
        inst_state->client_present--;
      }

      tmf8801_reval_instance_config(this, instance, state->sensor_type);
      inst_state->new_or_chng_req = false;
    }
  }
  else if(NULL != new_request)
  {
    if ((NULL == instance) && (new_request->message_id == SNS_STD_MSGID_SNS_STD_FLUSH_REQ))
    {
      // no op
    }
    else
    {
      // 1. If new request then:
      //     a. Power ON rails.
      //     b. Power ON COM port - Instance must handle COM port power.
      //     c. Create new instance.
      //     d. Re-evaluate existing requests and choose appropriate instance config.
      //     e. set_client_config for this instance.
      //     f. Add new_request to list of requests handled by the Instance.
      //     g. Power OFF COM port if not needed- Instance must handle COM port power.
      //     h. Return the Instance.
      // 2. If there is an Instance already present:
      //     a. Add new_request to list of requests handled by the Instance.
      //     b. Remove exist_request from list of requests handled by the Instance.
      //     c. Re-evaluate existing requests and choose appropriate Instance config.
      //     d. set_client_config for the Instance if not the same as current config.
      //     e. publish the updated config.
      //     f. Return the Instance.
      // 3.  If "flush" request:
      //     a. Perform flush on the instance.
      //     b. Return NULL.

      if(NULL == instance)
      {
        state->rail_config.rail_vote = SNS_RAIL_ON_NPM;
        flag = state->pwr_rail_service->api->sns_vote_power_rail_update(
                                             state->pwr_rail_service,
                                             this,
                                             &state->rail_config,
                                             &on_timestamp);
        if(flag == SNS_RC_SUCCESS){
            AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: no instance:  Turn on Rail");
        }
        else{
            AMS_PRINTF(HIGH, this, "AMS: ams_set_client_request: no instance:  rail turn on failed");
        }
        delta = sns_get_system_time() - on_timestamp;

        // Use on_timestamp to determine correct Timer value.
        if(delta < sns_convert_ns_to_ticks(AMS_DEVICE_OFF_TO_IDLE_MS*1000*1000))
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: reconfig power rail timer");
          tmf8801_start_power_rail_timer(this,
                                    sns_convert_ns_to_ticks(AMS_DEVICE_OFF_TO_IDLE_MS*1000*1000) - delta,
                                    AMS_POWER_RAIL_PENDING_SET_CLIENT_REQ);
        }
        else
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: power rail timer already configed");
          // rail is already ON
          reval_config = true;
        }

        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: null instance, call create_instance");

        /** create_instance() calls init() for the Sensor Instance */
        instance = this->cb->create_instance(this,
                                             sizeof(ams_instance_state));
        AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: create_instance SUCCESS");
        ams_instance_state *inst_state = (ams_instance_state *) instance->state->state;
        inst_state->new_or_chng_req = true;
        /* If rail is already ON then flag instance OK to configure */
        if(reval_config)
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: reval_config is TRUE, instance_is_ready_to_configure TRUE");
          inst_state->instance_is_ready_to_configure = true;
        }
        else
        {
            AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: reval_config is FALSE, instance_is_ready_to_configure FALSE");
            inst_state->instance_is_ready_to_configure = false;
        }
      }
      else
      {
        if(new_request->message_id == SNS_STD_MSGID_SNS_STD_FLUSH_REQ)
        {
           tmf8801_send_fifo_flush_done (instance, &state->my_suid);
           return instance;
        }
        else
        {
          reval_config = true;

          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: instance %p; new or chg rqst", instance);

          ams_instance_state *inst_state = (ams_instance_state *) instance->state->state;
          inst_state->new_or_chng_req = true;

          /** An existing client is changing request*/
          if((NULL != exist_request) && (NULL != new_request))
          {
            AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: chg rqst");
            instance->cb->remove_client_request(instance, exist_request);
          }
          /** A new client sent new_request*/
          else if(NULL != new_request)
          {
            // No-op. new_request will be added to requests list below.
          }
        }
      }

      /** Add the new request to list of client_requests.*/
      if(NULL != instance)
      {
        ams_instance_state *inst_state = (ams_instance_state*)instance->state->state;
        if((NULL != new_request) && (new_request->message_id != SNS_STD_MSGID_SNS_STD_FLUSH_REQ))
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: adding new request");
          instance->cb->add_client_request(instance, new_request);
          if(new_request->message_id ==
             SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_CONFIG)
          {
            AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: type PHYSICAL_SENSOR_TEST_CONFIG");

            if(ams_extract_self_test_info(this, instance, new_request))
            {
              AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: is selftest rqst; reval_config=%d, instance_is_ready_to_configure=%d"
              , reval_config, inst_state->instance_is_ready_to_configure);
              inst_state->new_self_test_request = true;
              inst_state->sensor_type_under_test = state->sensor_type;
            }
          } else {
            inst_state->client_present++;
          }
        }
        if(reval_config && inst_state->instance_is_ready_to_configure)
        {
          AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: call tmf8801_reval_instance_config");

          tmf8801_reval_instance_config(this, instance, state->sensor_type);
          if(inst_state->new_self_test_request)
          {
            tmf8801_set_self_test_inst_config(this, instance);
          }
        }
      }
    }
  }
  else // bad request
  {
    instance = NULL;
  }

  /* Sensors are required to call remove_instance when clientless */
  if((NULL != instance) &&
     (NULL == instance->cb->get_client_request(instance, p_prox_oc_suid, true)) &&
     (NULL == instance->cb->get_client_request(instance, p_prox_st_suid, true))
    )
  {
    sns_sensor *iter;
    ams_instance_state *inst_state = (ams_instance_state*)instance->state->state;
    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: NOW CLIENTLESS");
    for(iter = this->cb->get_library_sensor(this, true);
        NULL != iter;
        iter = this->cb->get_library_sensor(this, false))
    {
      ams_state *sensor_state = (ams_state*)iter->state->state;

      sensor_state->sensorSystemData = inst_state->instanceSystemData;
      // only update state version number when write to registry
      inst_state->cal_version[TOF_STATE_DATA_VERSION]++;
      sns_memscpy(sensor_state->cal_version,
                  sizeof(sensor_state->cal_version),
                  inst_state->cal_version,
                  sizeof(inst_state->cal_version));
    }

    tmf8801_write_calibration_to_registry(this);

    AMS_DEV_PRINTF(HIGH, this, "AMS: ams_set_client_request: no requests left, call cb->remove_instance");
    this->cb->remove_instance(instance);
  }

  return instance;
}
