/*
 * Copyright (c) 2019, ams AG
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>

#include "sns_tmf8801_sensor.h"
#include "sns_tmf8801_sensor_instance.h"


static void tmf8801_inst_cleanup(ams_instance_state *state, sns_stream_service *stream_mgr)
{
  if(NULL != state->interrupt_data_stream)
  {
    stream_mgr->api->remove_stream(stream_mgr, state->interrupt_data_stream);
    state->interrupt_data_stream = NULL;
  }
  if(NULL != state->timer_data_stream)
  {
    stream_mgr->api->remove_stream(stream_mgr, state->timer_data_stream);
    state->timer_data_stream = NULL;
  }
  if(NULL != state->scp_service)
  {
    state->scp_service->api->sns_scp_close(state->com_port_info.port_handle);
    state->scp_service->api->sns_scp_deregister_com_port(&state->com_port_info.port_handle);
    state->scp_service = NULL;
  }
}

static void tmf8801_init_diag_log_size(sns_sensor_instance *const this)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;

  float prox_temp[PROX_PROTO_STD_ARRAY_LEN] = {0.0, 0.0};

  uint8_t arr_index = 0;

  pb_float_arr_arg prox_arg = {.arr = (float*)prox_temp, .arr_len = PROX_PROTO_STD_ARRAY_LEN,
    .arr_index = &arr_index};

  uint64_t buffer[50];
  sns_memset(buffer, 0, sizeof(buffer));
  pb_ostream_t stream = pb_ostream_from_buffer((pb_byte_t *)buffer, sizeof(buffer));
  sns_diag_batch_sample batch_sample = sns_diag_batch_sample_init_default;

  batch_sample.sample.funcs.encode = &pb_encode_float_arr_cb;

  batch_sample.sample.arg = &prox_arg;
  if(pb_encode_tag(&stream, PB_WT_STRING,
                    sns_diag_sensor_state_raw_sample_tag))
  {
    if(pb_encode_delimited(&stream, sns_diag_batch_sample_fields,
                           &batch_sample))
    {
      state->prox_enc_raw_log_size = stream.bytes_written;
    }
  }
}

/** See sns_sensor_instance_api::init */
sns_rc tmf8801_inst_init(sns_sensor_instance *const this,
    sns_sensor_state const *sstate)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  ams_state *sensor_state = (ams_state*)sstate->state;
  float data[20];
  sns_service_manager *service_mgr = this->cb->get_service_manager(this);
  sns_stream_service *stream_mgr = (sns_stream_service*)
              service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);
  int64_t hwid = sensor_state->hardware_id;
  uint8_t thre;
  sns_rc rv = SNS_RC_SUCCESS;
  uint8_t rc;


  sns_sensor_uid * p_prox_oc_suid = (sns_sensor_uid *)tmf8801_get_suid_with_sensortype_and_hwid(AMS_PROX_OC, hwid);
  sns_sensor_uid * p_prox_st_suid = (sns_sensor_uid *)tmf8801_get_suid_with_sensortype_and_hwid(AMS_PROX_STRM, hwid);

  AMS_DEV_INST_PRINTF(LOW, this
                    , "AMS: tmf8801_inst_init: enter: this %p, sensor_state %p, sensor_type=%d"
                    , this
                    , sensor_state
                    , sensor_state->sensor_type
              );

  state->priv = sensor_state->priv; // local copy of our deviceCtx ptr
  state->sensor_type = sensor_state->sensor_type;
  state->publish_sensors = 0;
  state->publish_sensors_prev = 0;

  state->scp_service = (sns_sync_com_port_service*)
              service_mgr->get_service(service_mgr, SNS_SYNC_COM_PORT_SERVICE);

  state->diag_service =  (sns_diag_service*)
      service_mgr->get_service(service_mgr, SNS_DIAG_SERVICE);


  sns_memscpy (&state->irq_suid, sizeof(sns_sensor_uid),
        &sensor_state->irq_suid, sizeof(sns_sensor_uid));

  sns_memscpy (&state->timer_suid, sizeof(sns_sensor_uid),
        &sensor_state->timer_suid, sizeof(sns_sensor_uid));

  state->irq_info.irq_config = sensor_state->irq_config;


  /** Initialize COM port to be used by the Instance */
  sns_memscpy(&state->com_port_info.com_config,
              sizeof(state->com_port_info.com_config),
              &sensor_state->com_port_info.com_config,
              sizeof(sensor_state->com_port_info.com_config));

  state->scp_service->api->sns_scp_register_com_port(&state->com_port_info.com_config,
                                              &state->com_port_info.port_handle);
  AMS_DEV_INST_PRINTF(LOW, this
                    , "AMS: tmf8801_inst_init:  data_stream_:  scp handle %p"
                    , state->com_port_info.port_handle
              );
  if(NULL == state->com_port_info.port_handle)
  {
    tmf8801_inst_cleanup(state, stream_mgr);
    return SNS_RC_FAILED;
  }

  rv = state->scp_service->api->sns_scp_open(state->com_port_info.port_handle);
  if(SNS_RC_SUCCESS == rv)
  {
     AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_init: sns_scp_open SUCCESS");
     /* set our wrapped-up handle for this port */
     /* Fill out the instance's portHndl */
     state->port_handle.scp_service = state->scp_service;
     state->port_handle.scp_handle = state->com_port_info.port_handle;
  }
  else
  {
     AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_init: sns_scp_open failed");
  }

  /* Fill out the instance's portHndl */
  state->port_handle.diag = state->diag_service;
  state->port_handle_ptr = &state->port_handle;

  /**----------- Copy all Sensor UIDs in instance state -------------*/

  sns_memscpy(&state->prox_info.suid[AMS_OC_IDX],
              sizeof(state->prox_info.suid[AMS_OC_IDX]),
              p_prox_oc_suid,
              sizeof(state->prox_info.suid[AMS_OC_IDX]));
  sns_memscpy(&state->prox_info.suid[AMS_STRM_IDX],
              sizeof(state->prox_info.suid[AMS_STRM_IDX]),
              p_prox_st_suid,
              sizeof(state->prox_info.suid[AMS_STRM_IDX]));

  state->encoded_prox_event_len = pb_get_encoded_size_sensor_stream_event(data, PROX_PROTO_STD_ARRAY_LEN);

  state->instanceSystemData = sensor_state->sensorSystemData;
  sns_memscpy(state->cal_version,
              sizeof(state->cal_version),
              sensor_state->cal_version,
              sizeof(sensor_state->cal_version));

  // calculate acuracy according to threshold
  thre = (63 - sensor_state->sensorCmdData.threshold)/3;
  state->thresholdLow = sensor_state->sensorCmdData.threshold + thre;
  state->thresholdMed = sensor_state->sensorCmdData.threshold + (thre<<1);

  rc = tmf8801_wakeFromStandby(state->port_handle_ptr);
  if (!rc)
      return rc;
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_init: tmf8801_wakeFromStandby return True");

  rc = tof8801_app0WaitForIdle(state->port_handle_ptr);
  if (!rc)
      return rc;
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_init: tof8801_app0WaitForIdle return True");

  tof_deviceIdentifier_e chipId = AMS_UNKNOWN_DEVICE;
  chipId = tmf8801_validateDevice(state->port_handle_ptr);
  if (chipId != AMS_UNKNOWN_DEVICE){
    uint8_t TempByte[5];
    AMS_tmf8801_rwBlock(state->port_handle_ptr, TOF8801_REG_MAJ_VER, TempByte, 1, I2C_READ);
    AMS_tmf8801_rwBlock(state->port_handle_ptr, TOF8801_REG_MIN_VER, &TempByte[1], 4, I2C_READ);
    AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_init: App0 version: %u.%u.%u.%u\n", TempByte[0], TempByte[1],
                                    TempByte[2],TempByte[3]|(TempByte[4]<<8));
    tmf8801_deviceInit(state->priv, state->port_handle_ptr, &sensor_state->sensorSystemData, \
                 &sensor_state->sensorCmdData, chipId, hwid, TempByte);
  }
#ifdef AMS_DEV_DEBUG_ENABLE
        {
        uint8_t TempByte[5];
        AMS_tmf8801_rwBlock(state->port_handle_ptr, TOF8801_REG_APP_ID, TempByte, 3, I2C_READ);
        AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_init: ctx->appID = 0x%x, 0x%x, 0x%x", TempByte[0], TempByte[1],TempByte[2]);
        AMS_tmf8801_rwBlock(state->port_handle_ptr, TOF8801_REG_ENABLE, TempByte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_init: ctx->cpu_ready = 0x%x", TempByte[0]);
        }
#endif

  state->scp_service->api->sns_scp_update_bus_power(state->com_port_info.port_handle,
                                                                           false);

  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_init: close COM port");
  tmf8801_init_diag_log_size(this);
  return SNS_RC_SUCCESS;
}

bool tmf8801_validate_device_odr(sns_sensor_instance *const instance, float *sample_rate)
{
  ams_instance_state *state = (ams_instance_state*)instance->state->state;
  bool change_in_sr = false;
//#if CONFIG_AMS_USE_DRI - the test sensor is hardcoded to request 1 Hz so
//                         let's support it even in DRI mode
  if(*sample_rate > 0.0
     && *sample_rate <= AMS_DEVICE_ODR_1)
  {
    *sample_rate = AMS_DEVICE_ODR_1;
  }
  else if(*sample_rate > AMS_DEVICE_ODR_1
          && *sample_rate <= AMS_DEVICE_ODR_2)
  {
    *sample_rate = AMS_DEVICE_ODR_2;
  }
  else if(*sample_rate > AMS_DEVICE_ODR_2
          && *sample_rate <= AMS_DEVICE_ODR_5)
  {
    *sample_rate = AMS_DEVICE_ODR_5;
  }
  else if(*sample_rate > AMS_DEVICE_ODR_5
          && *sample_rate <= AMS_DEVICE_ODR_10)
  {
    *sample_rate = AMS_DEVICE_ODR_10;
  }
  else if(*sample_rate > AMS_DEVICE_ODR_10)
  {
    *sample_rate = AMS_DEVICE_ODR_20;
  }
  else
  {
    state->sampling_rate_hz = 0;
    state->sampling_intvl = 0;
  }

  if(state->sampling_rate_hz != *sample_rate)
  {
    change_in_sr = true;
    union {
        uint32_t  dwds[2];
        uint64_t qwd;
    } ticks;

    if(*sample_rate > 0)
    {
      state->sampling_intvl =
        sns_convert_ns_to_ticks((uint64_t)(1000000000.0 / *sample_rate));
      state->sampling_rate_hz = *sample_rate;
    }
    else
    {
      state->sampling_intvl = 0;
      state->sampling_rate_hz = 0.0f;
    }

    ticks.qwd = state->sampling_intvl;
    AMS_DEV_INST_PRINTF(LOW, instance
                    , "AMS: tmf8801_validate_device_odr: sample_rate= (%d), sampling_intvl=(%08x:%08x)"
                    , (uint32)*sample_rate
                    , ticks.dwds[1]
                    , ticks.dwds[0]
                    );
  }

  return change_in_sr;
}

/**
 * Runs a communication test - verfies WHO_AM_I, publishes self
 * test event.
 *
 * @param[i] instance  Instance reference
 * @param[i] uid       Sensor UID
 *
 * @return none
 */

static void _tmf8801_send_com_test_event(sns_sensor_instance *instance,
                                     sns_sensor_uid *uid, bool test_result)
{
  uint8_t data[1] = {0};
  pb_buffer_arg buff_arg = (pb_buffer_arg)
      { .buf = &data, .buf_len = sizeof(data) };
  sns_physical_sensor_test_event test_event =
     sns_physical_sensor_test_event_init_default;

  test_event.test_passed = test_result;
  test_event.test_type = SNS_PHYSICAL_SENSOR_TEST_TYPE_COM;
  test_event.test_data.funcs.encode = &pb_encode_string_cb;
  test_event.test_data.arg = &buff_arg;

  pb_send_event(instance,
                sns_physical_sensor_test_event_fields,
                &test_event,
                sns_get_system_time(),
                SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_EVENT,
                uid);
}

static void _tmf8801_setup_clean_irq_stream (sns_sensor_instance *this, bool on_of)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  sns_service_manager *service_mgr = this->cb->get_service_manager(this);
  sns_stream_service *stream_mgr = (sns_stream_service*)
              service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);

  if (on_of) {
    stream_mgr->api->create_sensor_instance_stream(stream_mgr,
                                                   this,
                                                   state->irq_suid,
                                                   &state->interrupt_data_stream);

    /** Initialize IRQ info to be used by the Instance */
    sns_data_stream* data_stream = state->interrupt_data_stream;
    uint8_t buffer[20];
    sns_request irq_req =
    {
      .message_id = SNS_INTERRUPT_MSGID_SNS_INTERRUPT_REQ,
      .request    = buffer
    };

    sns_memset(buffer, 0, sizeof(buffer));
    irq_req.request_len = pb_encode_request(buffer,
                                            sizeof(buffer),
                                            &state->irq_info.irq_config,
                                            sns_interrupt_req_fields,
                                            NULL);
    if(irq_req.request_len > 0)
    {
      data_stream->api->send_request(data_stream, &irq_req);
    }

    AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_init: sent IRQ rqst");

  } else {
    if(NULL != state->interrupt_data_stream)
    {
      tmf8801_deviceSetConfig(this, state->priv, AMS_CONFIG_GLOBAL, AMS_CONFIG_ALLOW_INTS, 0);
      stream_mgr->api->remove_stream(stream_mgr, state->interrupt_data_stream);
      state->interrupt_data_stream = NULL;
    }
  }
}

/**
 * Starts/restarts polling timer
 *
 * @param instance   Instance reference
 */

static void _tmf8801_start_polling_timer(sns_sensor_instance *this)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  sns_service_manager *service_mgr = this->cb->get_service_manager(this);
  sns_stream_service *stream_mgr = (sns_stream_service*)
              service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);

  sns_timer_sensor_config req_payload = sns_timer_sensor_config_init_default;
  uint8_t buffer[50];
  sns_request timer_req = {
    .message_id = SNS_TIMER_MSGID_SNS_TIMER_SENSOR_CONFIG,
    .request    = buffer
  };

  if (NULL == state->timer_data_stream)
  {
    stream_mgr->api->create_sensor_instance_stream(stream_mgr,
                                                   this,
                                                   state->timer_suid,
                                                   &state->timer_data_stream);
  }

  req_payload.is_periodic = true;
  req_payload.start_time = sns_get_system_time() - state->sampling_intvl + SNS_TIMER_LATENCY;
  req_payload.timeout_period = state->sampling_intvl;
  req_payload.start_config.early_start_delta = 0;

  timer_req.request_len = pb_encode_request(buffer, sizeof(buffer), &req_payload,
                                            sns_timer_sensor_config_fields, NULL);
  if(timer_req.request_len > 0)
  {
    state->timer_data_stream->api->send_request(state->timer_data_stream, &timer_req);
    state->timer_is_active = true;
  }
  else
  {
    AMS_INST_PRINTF(ERROR, this,"AMS: _tmf8801_start_polling_timer: LSM timer req encode error");
  }
}


static void _tmf8801_run_self_test(sns_sensor_instance *instance)
{
  ams_instance_state *state = (ams_instance_state*)instance->state->state;
  ams_deviceCtx_t *devCtx = (ams_deviceCtx_t *)(state->priv);
  bool who_am_i_success = 0;
  bool bret = 0;
  bool handlerIsReady = (devCtx && devCtx->portHndl);

  if(handlerIsReady)
  {
    tof_deviceIdentifier_e devId;
    devId = tmf8801_validateDevice(devCtx->portHndl);
    who_am_i_success = (devId != AMS_UNKNOWN_DEVICE) && (devId != AMS_LAST_DEVICE) ? true : false;
  }
  else
  {
    who_am_i_success = false;
  }

  if(state->prox_info.test_info[AMS_OC_IDX].test_client_present)
  {
    if(state->prox_info.test_info[AMS_OC_IDX].test_type == SNS_PHYSICAL_SENSOR_TEST_TYPE_COM)
    {
      _tmf8801_send_com_test_event(instance, &state->prox_info.suid[AMS_OC_IDX], who_am_i_success);
      state->new_self_test_request = false;
      state->prox_info.test_info[AMS_OC_IDX].test_client_present = false;
    }
    else if(state->prox_info.test_info[AMS_OC_IDX].test_type == SNS_PHYSICAL_SENSOR_TEST_TYPE_FACTORY)
    {
      AMS_DEV_INST_PRINTF(HIGH, instance,"AMS: _tmf8801_run_self_test: calling tmf8801_deviceCalibrate");
      bret = false;
      if(handlerIsReady)
      {
        bret = tmf8801_deviceCalibrate(instance, devCtx);
      }
#ifdef AMS_DEV_DEBUG_ENABLE
    {
        uint8_t tempbyte;
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_run_self_test: tofState is %d\n", devCtx->tofState);
        AMS_tmf8801_rwBlock(devCtx->portHndl, TOF8801_REG_INT_EN, &tempbyte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_run_self_test: device interrupt enabled is 0x%x\n", tempbyte);
    }
#endif
      if(!bret)
      {
        _tmf8801_send_factory_test_event(instance, &state->prox_info.suid[AMS_OC_IDX], false);
      }
      else
      {
        //_tmf8801_send_cal_event (instance, pSuid, devCtx->calDataFloat.calData, TMF8801_CALIBRATION_DATA_SIZE, NULL, 0);
      }
    }
    // we do not support TEST_HW for TMF8801
  }
  if(state->prox_info.test_info[AMS_STRM_IDX].test_client_present)
  {
    if(state->prox_info.test_info[AMS_STRM_IDX].test_type == SNS_PHYSICAL_SENSOR_TEST_TYPE_COM)
    {
      _tmf8801_send_com_test_event(instance, &state->prox_info.suid[AMS_STRM_IDX], who_am_i_success);
      state->new_self_test_request = false;
      state->prox_info.test_info[AMS_STRM_IDX].test_client_present = false;
    }
    else if(state->prox_info.test_info[AMS_STRM_IDX].test_type == SNS_PHYSICAL_SENSOR_TEST_TYPE_FACTORY)
    {
      AMS_DEV_INST_PRINTF(HIGH, instance,
                     "AMS: _tmf8801_run_self_test: calling tmf8801_deviceCalibrate");
      bret = false;
      if(handlerIsReady)
      {
        bret = tmf8801_deviceCalibrate(instance, devCtx);
      }
#ifdef AMS_DEV_DEBUG_ENABLE
    {
        uint8_t tempbyte;
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_run_self_test: tofState is %d\n", devCtx->tofState);
        AMS_tmf8801_rwBlock(devCtx->portHndl, TOF8801_REG_INT_EN, &tempbyte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_run_self_test: device interrupt enabled is 0x%x\n", tempbyte);
    }
#endif
      if (bret)
      {
        _tmf8801_start_polling_timer(instance);
        //_tmf8801_send_cal_event (instance, pSuid, devCtx->calDataFloat.calData, TMF8801_CALIBRATION_DATA_SIZE, NULL, 0);
      }
      else
      {
        _tmf8801_send_factory_test_event(instance, &state->prox_info.suid[AMS_STRM_IDX], false);
      }
    }
  }
}

/**
 * Updates polling configuration
 *
 * @param[i] instance   Sensor instance
 *
 * @return sampling interval time in ticks
 */

static void _tmf8801_set_polling_config(sns_sensor_instance *const this)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;

  {
    union {
        uint32_t dwds[2];
        uint64_t qwd;
    } ticks;
    ticks.qwd = state->sampling_intvl;
    AMS_DEV_INST_PRINTF(LOW, this
                    , "AMS: ams_set_polling_config: sampling_intvl=(%08x:%08x), timer_is_active=%d"
                    , ticks.dwds[1], ticks.dwds[0]
                    , state->timer_is_active
                    );
  }

  if(state->sampling_intvl > 0
     &&
     !state->new_self_test_request) // don't start the timer if self_test_request
  {
    _tmf8801_start_polling_timer(this);
  }
  else if(state->timer_is_active)
  {
    state->timer_is_active = false;

    sns_service_manager *service_mgr = this->cb->get_service_manager(this);
    sns_stream_service *stream_mgr = (sns_stream_service*)
                service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);
    if(NULL != state->timer_data_stream)
    {
      stream_mgr->api->remove_stream(stream_mgr, state->timer_data_stream);
      state->timer_data_stream = NULL;
    }
  }
}

/**
 * Configures sensor with new/recomputed settings
 *
 * @param instance   Instance reference
 */

static void _tmf8801_reconfig_hw(sns_sensor_instance *this)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;

  AMS_DEV_INST_PRINTF(LOW, this
                  , "AMS: _tmf8801_reconfig_hw: publish_sensors=0x%04x"
                  , state->publish_sensors
                  );

  if(state->publish_sensors & AMS_STREAMERS)
  {
    _tmf8801_set_polling_config(this);
  }
  else
  {
    /* QC - the periodic timer would still be running at this point; must close timer stream to turn off this timer */
    sns_service_manager *service_mgr = this->cb->get_service_manager(this);
    sns_stream_service *stream_mgr = (sns_stream_service*)
                service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);
    if(NULL != state->timer_data_stream)
    {
      stream_mgr->api->remove_stream(stream_mgr, state->timer_data_stream);
      state->timer_data_stream = NULL;
    }
    state->timer_is_active = false;
  }
}

/**
 * Sends config update event for the chosen sample_rate
 *
 * @param[i] instance    reference to this Instance
 */
static void _tmf8801_send_config_event(sns_sensor_instance *const instance)
{
  ams_instance_state *state =
     (ams_instance_state*)instance->state->state;
  ams_deviceCtx_t *devCtx = (ams_deviceCtx_t *)(state->priv);

  AMS_DEV_INST_PRINTF(LOW, instance
                  , "AMS: _tmf8801_send_config_event: enter: inst_this %p, .publish_sensors=0x%04x"
                  , instance
                  , state->publish_sensors
                  );

  if(state->publish_sensors & (AMS_PROX_OC|AMS_PROX_STRM))
  {
    sns_std_sensor_physical_config_event phy_sensor_config =
       sns_std_sensor_physical_config_event_init_default;
    // TODO: Use appropriate op_mode selected by driver.
    char operating_mode[] = "NORMAL";
    pb_buffer_arg op_mode_args;
    op_mode_args.buf = &operating_mode[0];
    op_mode_args.buf_len = sizeof(operating_mode);

    ams_deviceInfo_t info;
    AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: _tmf8801_send_config_event calling tmf8801_getDeviceInfo with hwid = %d" , devCtx->hwid);
    tmf8801_getDeviceInfo(&info, devCtx->hwid);

    if(state->publish_sensors & AMS_PROX_OC)
    {
      phy_sensor_config.has_sample_rate = false;
      phy_sensor_config.has_dri_enabled = true;
      phy_sensor_config.dri_enabled = true;

      phy_sensor_config.has_water_mark = false;
      phy_sensor_config.has_active_current = true;
      phy_sensor_config.operation_mode.funcs.encode = &pb_encode_string_cb;
      phy_sensor_config.operation_mode.arg = &op_mode_args;
      phy_sensor_config.active_current = info.tofSensor.activeCurrent_uA;
      phy_sensor_config.range_count = 2;
      phy_sensor_config.range[0] = info.tofSensor.rangeMin;
      phy_sensor_config.range[1] = info.tofSensor.rangeMax;
      phy_sensor_config.has_stream_is_synchronous = false;
      phy_sensor_config.has_DAE_watermark = false;

      AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_send_config_event: send PROX_OC cfg evt");
      pb_send_event(instance,
                    sns_std_sensor_physical_config_event_fields,
                    &phy_sensor_config,
                    sns_get_system_time(),
                    SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_PHYSICAL_CONFIG_EVENT,
                    &state->prox_info.suid[AMS_OC_IDX]);
      if (state->prox_info.client_cfg.oc_send_last_event)
      {
        state->prox_info.client_cfg.oc_send_last_event = false;
        AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_send_config_event: send PROX_OC last data evt");
        pb_send_event(instance,
                      sns_proximity_event_fields,
                      &state->prox_info.prox_state,
                      state->prox_info.timestamp,
                      SNS_PROXIMITY_MSGID_SNS_PROXIMITY_EVENT,
                      &state->prox_info.suid[AMS_OC_IDX]);
      }
    }
    if(state->publish_sensors & AMS_PROX_STRM)
    {
      phy_sensor_config.has_sample_rate = true;
      phy_sensor_config.sample_rate = state->sampling_rate_hz;
      phy_sensor_config.has_dri_enabled = true;
      phy_sensor_config.dri_enabled = false;

      phy_sensor_config.has_water_mark = false;
      phy_sensor_config.has_active_current = true;
      phy_sensor_config.operation_mode.funcs.encode = &pb_encode_string_cb;
      phy_sensor_config.operation_mode.arg = &op_mode_args;
      phy_sensor_config.active_current = info.tofSensor.activeCurrent_uA;
      phy_sensor_config.range_count = 2;
      phy_sensor_config.range[0] = info.tofSensor.rangeMin;
      phy_sensor_config.range[1] = info.tofSensor.rangeMax;
      phy_sensor_config.has_DAE_watermark = false;

      AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_send_config_event: send PROX_STRM cfg evt");

      pb_send_event(instance,
                    sns_std_sensor_physical_config_event_fields,
                    &phy_sensor_config,
                    sns_get_system_time(),
                    SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_PHYSICAL_CONFIG_EVENT,
                    &state->prox_info.suid[AMS_STRM_IDX]);
    }
  }
}

/** See sns_sensor_instance_api::set_client_config */
sns_rc tmf8801_inst_set_client_config(sns_sensor_instance *const this,
    sns_request const *client_request)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  float desired_sample_rate = 0.0;
  float desired_report_rate = 0.0;
  float ams_chosen_sample_rate = 0.0;
  sns_rc rv = SNS_RC_SUCCESS;
#ifdef AMS_DEV_DEBUG_ENABLE
  ams_deviceCtx_t *devCtx = (ams_deviceCtx_t *)(state->priv);
#endif
  sns_std_req *payload = (sns_std_req*)client_request->request;

  AMS_DEV_INST_PRINTF(LOW, this
                  , "AMS: tmf8801_inst_set_client_config: enter: inst_this 0x%X, message_id %d"
                  , this, client_request->message_id);

  // Turn COM port ON
  state->scp_service->api->sns_scp_update_bus_power(state->com_port_info.port_handle,
                                                                           true);
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_set_client_config: open COM port");

#ifdef AMS_DEV_DEBUG_ENABLE
        {
        uint8_t TempByte[5];
        AMS_tmf8801_rwBlock(devCtx->portHndl, TOF8801_REG_APP_ID, TempByte, 3, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_set_client_config: ctx->appID = 0x%x, 0x%x, 0x%x", TempByte[0], TempByte[1],TempByte[2]);
        AMS_tmf8801_rwBlock(devCtx->portHndl, TOF8801_REG_ENABLE, TempByte, 1, I2C_READ);
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_set_client_config: ctx->cpu_ready = 0x%x", TempByte[0]);
        }
#endif

  if(client_request->message_id == SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_CONFIG)
  {
    AMS_DEV_INST_PRINTF(LOW, this
                    , "AMS: tmf8801_inst_set_client_config: _STD_SENSOR_CONFIG, SR=%d, RR=%d"
                    , (int32_t)payload->desired_sample_rate
                    , (int32_t)payload->desired_report_rate
                    );

    // 1. Extract sample, report rates from client_request.
    // 2. Configure sensor HW.
    // 3. sendRequest() for Timer to start/stop in case of polling using timer_data_stream.
    // 4. sendRequest() for Intrerupt register/de-register in case of DRI using interrupt_data_stream.
    // 5. Save the current config information like type, sample_rate, report_rate, etc.
    desired_sample_rate = payload->desired_sample_rate;
    desired_report_rate = payload->desired_report_rate;

    if(desired_report_rate > desired_sample_rate)
    {
      // bad request. Return error or default report_rate to sample_rate
      desired_report_rate = desired_sample_rate;
    }

    ams_chosen_sample_rate = desired_sample_rate; //we only put this into effect in polling mode
    AMS_DEV_INST_PRINTF(LOW, this
                    , "AMS: tmf8801_inst_set_client_config: ams_chosen_sample_rate=%d"
                    , ams_chosen_sample_rate
                    );

    {
      ams_configureFeature_t feature;
      uint16_t EITHER = 0;
      bool     both_off = false, both_were_off = false;
      uint16_t newly_on = 0, newly_off = 0;
      bool     option_changed = false;
      bool option_new_state = false;
      uint16_t on = state->publish_sensors;
      uint16_t p = state->publish_sensors_prev;

      newly_on = ~p & on;
      newly_off = p & ~on;

      AMS_DEV_INST_PRINTF(LOW, this
                      , "AMS: tmf8801_inst_set_client_config: on=%04x, p=%04x, newly_on=%04x, newly_off=%04x"
                      , (uint32_t)on
                      , (uint32_t)p
                      , (uint32_t)newly_on
                      , (uint32_t)newly_off
                      );

      feature = AMS_CONFIG_PROX;
      EITHER = AMS_PROX_OC | AMS_PROX_STRM;
      both_were_off = !(p & EITHER);
      both_off = !(on & EITHER);

      AMS_DEV_INST_PRINTF(LOW, this
                      , "AMS: tmf8801_inst_set_client_config: PROX: EITHER=%04x, both_were_off=%d, both_off=%d"
                      , (uint32_t)EITHER
                      , (uint32_t)both_were_off
                      , (uint32_t)both_off
                      );

      option_changed = false;
      if (newly_off & AMS_PROX_STRM)
      {
        option_changed = true;
        option_new_state = 0;
      }
      else if (newly_on & AMS_PROX_STRM)
      {
        option_changed = true;
        option_new_state = 1;
      }
      if(option_changed && !state->new_self_test_request)
      {
        AMS_DEV_INST_PRINTF(LOW, this
                        , "AMS: tmf8801_inst_set_client_config: PROX: CHANGE ALWAYS_READ STATE TO:  %d"
                        , option_new_state
                        );
        tmf8801_deviceSetConfig(this, state->priv, feature, AMS_CONFIG_ALWAYS_READ, option_new_state);
      }

      option_changed = false;
      if ((newly_on & EITHER) && both_were_off)
      {
        option_changed = true;
        option_new_state = 1;
      }
      else if((newly_off & EITHER) && both_off)
      {
        option_changed = true;
        option_new_state = 0;
      }
      if (newly_on & AMS_PROX_OC)
      {
        state->prox_info.prev_OC_report[0] = AMS_PROXIMITY_EVENT_TYPE_INVALID;
        /* if we're not going to _CONFIG_ENABLE it, make the 1st OC report now */
        if(!option_changed)
        {
          log_sensor_state_raw_info log_prox_state_raw_info;
          // Allocate Sensor State Raw log packets
          sns_memzero(&log_prox_state_raw_info, sizeof(log_prox_state_raw_info));

          tmf8801_generateProxResponse(state, state->diag_service, sns_get_system_time(),
                                 this, &log_prox_state_raw_info, AMS_REPORT_ON_CHANGE);
        }
      }
      if(option_changed && !state->new_self_test_request)
      {
        AMS_DEV_INST_PRINTF(LOW, this
                        , "AMS: tmf8801_inst_set_client_config: PROX: CHANGE ENABLE STATE TO:  %d"
                        , option_new_state
                        );
        tmf8801_deviceSetConfig(this, state->priv, feature, AMS_CONFIG_ENABLE, option_new_state);
      }

      {
        uint16_t p_streamers = p & AMS_STREAMERS;
        uint16_t on_streamers = on & AMS_STREAMERS;

        /* Update timer if any change in current/previous streaming sensors */
        if(on_streamers || p_streamers)
        {
          if(tmf8801_validate_device_odr(this, &ams_chosen_sample_rate))
          {
             _tmf8801_reconfig_hw(this);
          }
        }
      }
      {
        uint16_t p_on_changesrs = p & AMS_ON_CHANGERS;
        uint16_t on_changesrs = on & AMS_ON_CHANGERS;
        if (!p_on_changesrs && on_changesrs)
        {
          // setup the irq stream
          _tmf8801_setup_clean_irq_stream (this, true);
        }
        if (p_on_changesrs && !on_changesrs)
        {
          // remove the irq stream
          _tmf8801_setup_clean_irq_stream (this, false);
        }
      }

      if (state->new_or_chng_req)
      {
        sns_sensor_uid *pSuid = NULL;
        if (newly_on & (AMS_PROX_OC | AMS_PROX_STRM))
        {
          if (newly_on & AMS_PROX_OC)
          {
            pSuid = &(state->prox_info.suid[AMS_OC_IDX]);
          }
          if (newly_on & AMS_PROX_STRM)
          {
            pSuid = &(state->prox_info.suid[AMS_STRM_IDX]);
          }
        }
        state->new_or_chng_req = false;
      }
    }

    state->publish_sensors_prev = state->publish_sensors;
    if(!state->new_self_test_request)
    {
      _tmf8801_send_config_event(this);
    }
  }
  else if(client_request->message_id == SNS_STD_MSGID_SNS_STD_FLUSH_REQ)
  {
    AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_set_client_config: _STD_FLUSH_REQ");
  }
  else if(client_request->message_id ==
          SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_CONFIG)
  {
    AMS_DEV_INST_PRINTF(MED, this
                    , "AMS: tmf8801_inst_set_client_config: _PHYSICAL_SENSOR_TEST_CONFIG"
                    );

     // 1. Extract test type from client_request.
     // 2. Configure sensor HW for test type.
     // 3. send_request() for Timer Sensor in case test needs polling/waits.
     // 4. Factory test is TBD.
     _tmf8801_run_self_test(this);

  }

  // Turn COM port OFF
  state->scp_service->api->sns_scp_update_bus_power(state->com_port_info.port_handle,
                                                                           false);
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_set_client_config: close COM port");
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_set_client_config: end");

  return rv;
}

sns_rc tmf8801_inst_deinit(sns_sensor_instance *const this)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  sns_service_manager *service_mgr = this->cb->get_service_manager(this);
  sns_stream_service *stream_mgr =
      (sns_stream_service*)service_mgr->get_service(service_mgr,
                                                    SNS_STREAM_SERVICE);

  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_deinit");

  tmf8801_inst_cleanup(state, stream_mgr);

  return SNS_RC_SUCCESS;
}
