/*
 * Copyright (c) 2019, ams AG
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "sns_tmf8801_sensor.h"
#include "sns_tmf8801_sensor_instance.h"

static void _tmf8801_wrap_up_factory_test(sns_sensor_instance *instance,
                                         bool *test_result);
static void _tmf8801_check_and_remove_stream(sns_sensor_instance *this,
    ams_mode_index oc_strm, sns_data_stream *stream);

/**
 * Allocate Sensor State Raw Log Packet
 *
 * @param[i] diag       Pointer to diag service
 * @param[i] log_size   Optional size of log packet to
 *    be allocated. If not provided by setting to 0, will
 *    default to using maximum supported log packet size
 */
static void _tmf8801_log_sensor_state_raw_alloc(
  log_sensor_state_raw_info *log_raw_info,
  uint32_t log_size)
{
  uint32_t max_log_size = 0;

  if(NULL == log_raw_info || NULL == log_raw_info->diag ||
     NULL == log_raw_info->instance || NULL == log_raw_info->sensor_uid)
  {
    return;
  }

  // allocate memory for sensor state - raw sensor log packet
  max_log_size = log_raw_info->diag->api->get_max_log_size(
       log_raw_info->diag);

  if(0 == log_size)
  {
    // log size not specified.
    // Use max supported log packet size
    log_raw_info->log_size = max_log_size;
  }
  else if(log_size <= max_log_size)
  {
    log_raw_info->log_size = log_size;
  }
  else
  {
    return;
  }

  log_raw_info->log = log_raw_info->diag->api->alloc_log(
    log_raw_info->diag,
    log_raw_info->instance,
    log_raw_info->sensor_uid,
    log_raw_info->log_size,
    SNS_DIAG_SENSOR_STATE_LOG_RAW);

  log_raw_info->log_sample_cnt = 0;
  log_raw_info->bytes_written = 0;
}

static sns_rc tmf8801_encode_prox_log_sensor_state_raw(
  void *log, size_t log_size, size_t encoded_log_size, void *encoded_log,
  size_t *bytes_written)
{
  sns_rc rc = SNS_RC_SUCCESS;
  uint32_t i = 0;
  size_t encoded_sample_size = 0;
  size_t parsed_log_size = 0;
  sns_diag_batch_sample batch_sample = sns_diag_batch_sample_init_default;
  uint8_t arr_index = 0;
  float temp[PROX_PROTO_STD_ARRAY_LEN];
  pb_float_arr_arg arg = {.arr = (float *)temp, .arr_len = PROX_PROTO_STD_ARRAY_LEN,
    .arr_index = &arr_index};

  if(NULL == encoded_log || NULL == log || NULL == bytes_written)
  {
    return SNS_RC_FAILED;
  }

  batch_sample.sample.funcs.encode = &pb_encode_float_arr_cb;
  batch_sample.sample.arg = &arg;

  if(!pb_get_encoded_size(&encoded_sample_size, sns_diag_batch_sample_fields,
                          &batch_sample))
  {
    return SNS_RC_FAILED;
  }

  pb_ostream_t stream = pb_ostream_from_buffer(encoded_log, encoded_log_size);
  prox_batch_sample *raw_sample = (prox_batch_sample *)log;

  while(parsed_log_size < log_size &&
        (stream.bytes_written + encoded_sample_size)<= encoded_log_size &&
        i < (uint32_t)(-1))
  {
    arr_index = 0;
    arg.arr = (float *)raw_sample[i].sample;

    batch_sample.sample_type = raw_sample[i].sample_type;
    batch_sample.status = raw_sample[i].status;
    batch_sample.timestamp = raw_sample[i].timestamp;

    if(!pb_encode_tag(&stream, PB_WT_STRING,
                      sns_diag_sensor_state_raw_sample_tag))
    {
      rc = SNS_RC_FAILED;
      break;
    }
    else if(!pb_encode_delimited(&stream, sns_diag_batch_sample_fields,
                                 &batch_sample))
    {
      rc = SNS_RC_FAILED;
      break;
    }

    parsed_log_size += sizeof(prox_batch_sample);
    i++;
  }

  if (SNS_RC_SUCCESS == rc)
  {
    *bytes_written = stream.bytes_written;
  }

  return rc;
}

/**
 *
 * Add raw uncalibrated sensor data to Sensor State Raw log
 * packet
 *
 * @param[i] log_raw_info Pointer to logging information
 *                        pertaining to the sensor
 * @param[i] raw_data     Uncalibrated sensor data to be logged
 * @param[i] timestamp    Timestamp of the sensor data
 * @param[i] status       Status of the sensor data
 *
 * * @return sns_rc,
 * SNS_RC_SUCCESS if encoding was succesful
 * SNS_RC_FAILED otherwise
 */
static sns_rc _tmf8801_log_sensor_state_raw_add(
  log_sensor_state_raw_info *log_raw_info,
  float *raw_data,
  sns_time timestamp,
  sns_std_sensor_sample_status status)
{
  sns_rc rc = SNS_RC_SUCCESS;

  if(NULL == log_raw_info || NULL == log_raw_info->diag ||
     NULL == log_raw_info->instance || NULL == log_raw_info->sensor_uid ||
     NULL == raw_data || NULL == log_raw_info->log)
  {
    return SNS_RC_FAILED;
  }

  if(log_raw_info->sensor_type == AMS_PROX_OC
     ||
     log_raw_info->sensor_type == AMS_PROX_STRM)
  {
    prox_batch_sample *sample =
        (prox_batch_sample *)log_raw_info->log;

    if(0 == log_raw_info->batch_sample_cnt)
    {
      sample[log_raw_info->log_sample_cnt].sample_type =
        SNS_DIAG_BATCH_SAMPLE_TYPE_FIRST;
    }
    else
    {
      sample[log_raw_info->log_sample_cnt].sample_type =
        SNS_DIAG_BATCH_SAMPLE_TYPE_INTERMEDIATE;
    }

    sample[log_raw_info->log_sample_cnt].timestamp = timestamp;

    sns_memscpy(sample[log_raw_info->log_sample_cnt].sample,
                sizeof(sample[log_raw_info->log_sample_cnt].sample),
                raw_data,
                sizeof(sample[log_raw_info->log_sample_cnt].sample));

    sample[log_raw_info->log_sample_cnt].status = status;

    log_raw_info->bytes_written += sizeof(prox_batch_sample);

    log_raw_info->log_sample_cnt++;
    log_raw_info->batch_sample_cnt++;
  }
  return rc;
}

/**
 * Submit the Sensor State Raw Log Packet
 *
 * @param[i] log_raw_info   Pointer to logging information
 *       pertaining to the sensor
 * @param[i] batch_complete true if submit request is for end
 *       of batch
 *  */
static void _tmf8801_log_sensor_state_raw_submit(
  log_sensor_state_raw_info *log_raw_info,
  bool batch_complete)
{
  sns_diag_encode_log_cb cb;

  if(NULL == log_raw_info || NULL == log_raw_info->diag ||
     NULL == log_raw_info->instance || NULL == log_raw_info->sensor_uid ||
     NULL == log_raw_info->log)
  {
    return;
  }

  if(log_raw_info->sensor_type == AMS_PROX_OC
     ||
     log_raw_info->sensor_type == AMS_PROX_STRM)
  {
    prox_batch_sample *sample = (prox_batch_sample *)log_raw_info->log;

    if(batch_complete)
    {
      // overwriting previously sample_type for last sample
      if(1 == log_raw_info->batch_sample_cnt)
      {
        sample[0].sample_type =
          SNS_DIAG_BATCH_SAMPLE_TYPE_ONLY;
      }
      else if(1 < log_raw_info->batch_sample_cnt)
      {
        sample[log_raw_info->log_sample_cnt - 1].sample_type =
          SNS_DIAG_BATCH_SAMPLE_TYPE_LAST;
      }
    }
    cb = tmf8801_encode_prox_log_sensor_state_raw;
  }
  else
  {
    return;
  }

  log_raw_info->diag->api->submit_log(
        log_raw_info->diag,
        log_raw_info->instance,
        log_raw_info->sensor_uid,
        log_raw_info->bytes_written,
        log_raw_info->log,
        SNS_DIAG_SENSOR_STATE_LOG_RAW,
        log_raw_info->log_sample_cnt * log_raw_info->encoded_sample_size,
        cb);

  log_raw_info->log = NULL;
}


sns_std_sensor_sample_status tmf8801_getProxAccuracy(ams_instance_state *instance, ams_apiPrx_t prox_result )
{
    sns_std_sensor_sample_status accuracy;

    if(prox_result.reliability == 0)
    {
       accuracy = SNS_STD_SENSOR_SAMPLE_STATUS_UNRELIABLE;
    }
    else if(prox_result.distance <=200)
    {
        // for distance <= 200mm, confidence is either 1 or 10
        // 1 means not calibrated, 10 means calibrated
        if (prox_result.reliability < 5)
            accuracy = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_LOW;
        else
            accuracy = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_HIGH;
    }
    else
    {
        if (prox_result.reliability < instance->thresholdLow)
            accuracy = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_LOW;
        else if (prox_result.reliability < instance->thresholdMed)
            accuracy = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_MEDIUM;
        else
            accuracy = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_HIGH;
    }
    return accuracy;
}

static void tmf8801_getTofResult(void *dcbCtx, ams_apiPrx_t *result){
    ams_tof_capture_t *tofCapData = &((ams_deviceCtx_t *)dcbCtx)->tofCapData;

    result->distance = tofCapData->distance;
    result->reliability = tofCapData->reliability & RELIABILITY_MASK;
    result->timeStamp = tofCapData->timeStamp;
    if (result->distance > 200) {

        result->nearBy = false;
    }
    else{
        result->nearBy = true;
    }
}

/**
 * Wrap up a Proximity sample and generate an event.
 *
 * @param[i] state              The state of the sensor instance
 * @param[i] diag               The instance's diag_service
 * @param[i] timestamp          Timestamp to be used for this sample
 * @param[i] instance           The current sensor instance
 * @param[i] log_prox_state_raw_info  Info on the log packet
 */
void tmf8801_generateProxResponse(ams_instance_state *state,
                              sns_diag_service* diag,
                              sns_time timestamp,
                              sns_sensor_instance *const instance,
                              log_sensor_state_raw_info *log_prox_state_raw_info,
                              ams_reportType_t report_type)
{
  UNUSED_VAR(diag);
  log_prox_state_raw_info->encoded_sample_size = state->prox_enc_raw_log_size;
  log_prox_state_raw_info->diag = diag;
  log_prox_state_raw_info->instance = instance;

  ams_apiPrx_t outData;
  ams_proximity_event_type ams_newNearBy;
  sns_std_sensor_sample_status status = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_HIGH;
  float data[PROX_PROTO_STD_ARRAY_LEN];

  tmf8801_getTofResult(state->priv, &outData);
  status = tmf8801_getProxAccuracy(state, outData);

  AMS_DEV_INST_PRINTF(HIGH, instance, "AMS: tmf8801_generateProxResponse: distance=%d, nearBy=%d", outData.distance, outData.nearBy);

  state->prox_info.prox_state.raw_adc = (outData.distance&0xFFFF) | ((outData.reliability&0xFF) << 24);
  state->prox_info.prox_state.status = status;
  /* The event type code conveys the binary nearBy status */
  state->prox_info.prox_state.proximity_event_type =
         (outData.nearBy ? SNS_PROXIMITY_EVENT_TYPE_NEAR : SNS_PROXIMITY_EVENT_TYPE_FAR);
  ams_newNearBy = outData.nearBy ? AMS_PROXIMITY_EVENT_TYPE_NEAR : AMS_PROXIMITY_EVENT_TYPE_FAR;

  data[0] = (float)(outData.distance);
  data[1] = (float)(outData.reliability);

  if((state->publish_sensors & AMS_PROX_OC) &&
     ((report_type == AMS_REPORT_ON_CHANGE) || (report_type == AMS_REPORT_OC_AND_STRM)))
  {
    if (ams_newNearBy != state->prox_info.prev_OC_report[0])
    {
      pb_send_event(instance,
                    sns_proximity_event_fields,
                    &state->prox_info.prox_state,
                    timestamp,
                    SNS_PROXIMITY_MSGID_SNS_PROXIMITY_EVENT,
                    &state->prox_info.suid[AMS_OC_IDX]);

      /* For on-change only, cache result data in the instance state.
       * This is so the framework can publish the cached data
       * to multiple clients.
       */
      state->prox_info.prev_OC_report[0] = ams_newNearBy;
      state->prox_info.timestamp = timestamp;
#ifdef CONFIG_PROX_INCLUDE_PROXIMITY_RESULTS
      pb_send_sensor_stream_event(instance,
                                &state->prox_info.suid[AMS_OC_IDX],
                                timestamp,
                                SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_EVENT,
                                status,
                                data,
                                ARR_SIZE(data),
                                state->encoded_prox_event_len);
#endif
        log_prox_state_raw_info->sensor_uid = &state->prox_info.suid[AMS_OC_IDX];
        log_prox_state_raw_info->sensor_type = AMS_PROX_OC;

        _tmf8801_log_sensor_state_raw_alloc(log_prox_state_raw_info, sizeof(prox_batch_sample));
        _tmf8801_log_sensor_state_raw_add(
            log_prox_state_raw_info,
            data,
            timestamp,
            status);
        _tmf8801_log_sensor_state_raw_submit(log_prox_state_raw_info, true);
    }
  }
  if((state->publish_sensors & AMS_PROX_STRM) &&
     ((report_type == AMS_REPORT_STREAMING) || (report_type == AMS_REPORT_OC_AND_STRM)))
  {
    pb_send_event(instance,
                  sns_proximity_event_fields,
                  &state->prox_info.prox_state,
                  timestamp,
                  SNS_PROXIMITY_MSGID_SNS_PROXIMITY_EVENT,
                  &state->prox_info.suid[AMS_STRM_IDX]);
#ifdef CONFIG_PROX_INCLUDE_PROXIMITY_RESULTS
    pb_send_sensor_stream_event(instance,
                                 &state->prox_info.suid[AMS_STRM_IDX],
                                 timestamp,
                                 SNS_STD_SENSOR_MSGID_SNS_STD_SENSOR_EVENT,
                                 status,
                                 data,
                                 ARR_SIZE(data),
                                 state->encoded_prox_event_len);
#endif
    log_prox_state_raw_info->sensor_uid = &state->prox_info.suid[AMS_STRM_IDX];
    log_prox_state_raw_info->sensor_type = AMS_PROX_STRM;

    _tmf8801_log_sensor_state_raw_alloc(log_prox_state_raw_info, sizeof(prox_batch_sample));
    _tmf8801_log_sensor_state_raw_add(
        log_prox_state_raw_info,
        data,
        timestamp,
        status);
    _tmf8801_log_sensor_state_raw_submit(log_prox_state_raw_info, true);
  }
}

/** See sns_sensor_instance_api::notify_event */
static sns_rc tmf8801_inst_notify_event(sns_sensor_instance *const this)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  sns_interrupt_event irq_event = sns_interrupt_event_init_zero;
  sns_sensor_event *event;
  sns_diag_service* diag = state->diag_service;
  ams_deviceCtx_t *devCtx = (ams_deviceCtx_t *)(state->priv);
  int32_t updateAvailableFlags;
  bool handlerIsReady = (devCtx && devCtx->portHndl && devCtx->portHndl->scp_handle);

  AMS_DEV_INST_PRINTF(LOW, this
                  , "AMS: tmf8801_inst_notify_event: enter, data_streams: int %p, timer %p"
                  , state->interrupt_data_stream
                  , state->timer_data_stream
                  );

  // Turn COM port ON
  state->scp_service->api->sns_scp_update_bus_power(state->com_port_info.port_handle,true);
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: open COM port");
  // Handle interrupts
  updateAvailableFlags = 0;
  if(NULL != state->interrupt_data_stream)
  {
    event = state->interrupt_data_stream->api->peek_input(state->interrupt_data_stream);
    while(NULL != event)
    {
      AMS_DEV_INST_PRINTF(LOW, this
                      , "AMS: tmf8801_inst_notify_event: next interrupt event:  id=%d"
                      , event->message_id
                      );
      if (event->message_id == SNS_INTERRUPT_MSGID_SNS_INTERRUPT_REG_EVENT)
      {
        if (state->priv)
        {
          AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: Received INT registration event");
          tmf8801_deviceSetConfig(this, state->priv, AMS_CONFIG_GLOBAL, AMS_CONFIG_ALLOW_INTS, 1);
        }
      }
      else if (event->message_id != SNS_INTERRUPT_MSGID_SNS_INTERRUPT_EVENT)
      {
         AMS_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_notify_event: Received non-rupt INT event id=%d", event->message_id);
      }
      else
      {
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: Received  INT event, sensors are (0x%x)",state->publish_sensors );
        pb_istream_t stream = pb_istream_from_buffer((pb_byte_t*)event->event,
                                                     event->event_len);
        if(state->publish_sensors & AMS_ON_CHANGERS)
        {
          AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: Received valid INT event");
          if(pb_decode(&stream, sns_interrupt_event_fields, &irq_event))
          {
            state->interrupt_timestamp = irq_event.timestamp;
            if (handlerIsReady){
              tmf8801_deviceEventHandler(this, devCtx);
              updateAvailableFlags = tmf8801_getUpdateResult(state->priv);
              AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: updateAvailableFlags=%08x", updateAvailableFlags);
              if (updateAvailableFlags == AMS_CALIBRATION_DONE)
              {
                bool test_result = false;

                if (updateAvailableFlags == AMS_CALIBRATION_DONE){
                  test_result = true;
                }
                _tmf8801_wrap_up_factory_test(this, &test_result);
              }
              else if(!state->new_self_test_request)
              {
                /* no regular reports if selftesting (including if CAL just completed) */
                if (updateAvailableFlags & (1 << AMS_PROXIMITY_SENSOR)){
                  AMS_DEV_INST_PRINTF(HIGH, this, "AMS: tmf8801_inst_notify_event: updateAvailable => AMS_PROXIMITY_SENSOR");
                  if(state->publish_sensors & AMS_PROX_OC){
                    log_sensor_state_raw_info log_prox_state_raw_info;

                    // Allocate Sensor State Raw log packets
                    sns_memzero(&log_prox_state_raw_info, sizeof(log_prox_state_raw_info));

                    tmf8801_generateProxResponse(state, diag, state->interrupt_timestamp,
                                           this, &log_prox_state_raw_info, AMS_REPORT_ON_CHANGE);
                  }
                }
              }
            }
          }
          else
          {
            AMS_INST_PRINTF(ERROR, this, "AMS: tmf8801_inst_notify_event: pb_decode error");
          }
        }
      }
      event = state->interrupt_data_stream->api->get_next_input(state->interrupt_data_stream);
    }
    if (updateAvailableFlags == AMS_CALIBRATION_DONE)
    {
      _tmf8801_check_and_remove_stream(this, AMS_OC_IDX, state->interrupt_data_stream);
    }

    AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: no more selftest timer events");
  }

  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: start of timer handler");
  // Handle Timer events
  updateAvailableFlags = 0;
  if(NULL != state->timer_data_stream)
  {
    event = state->timer_data_stream->api->peek_input(state->timer_data_stream);
    while(NULL != event)
    {
      pb_istream_t stream = pb_istream_from_buffer((pb_byte_t*)event->event,
                                                     event->event_len);
      sns_timer_sensor_event timer_event;

      AMS_DEV_INST_PRINTF(LOW, this
                      , "AMS: tmf8801_inst_notify_event: next timer event:  id=%d; timer_is_active=%d"
                      , event->message_id
                      , state->timer_is_active
                      );

      if(pb_decode(&stream, sns_timer_sensor_event_fields, &timer_event))
      {
        if((event->message_id == SNS_TIMER_MSGID_SNS_TIMER_SENSOR_EVENT)
            &&
            state->timer_is_active
            &&
            state->sampling_intvl > 0)
        {
          AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: poll timer: publish_sensors 0x%04x"
                                       , state->publish_sensors
                                       );
          if (handlerIsReady)
          {
            tmf8801_deviceEventHandler(this, devCtx);
            updateAvailableFlags = tmf8801_getUpdateResult(state->priv);
            AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: updateAvailableFlags=%08x", updateAvailableFlags);

            if (updateAvailableFlags == AMS_CALIBRATION_DONE)
            {
                bool test_result = false;

                if (updateAvailableFlags == AMS_CALIBRATION_DONE){
                  test_result = true;
                }
                _tmf8801_wrap_up_factory_test(this, &test_result);
            }
            else if(!state->new_self_test_request)
            {
                if(state->publish_sensors & AMS_PROX_STRM)
                {
                  log_sensor_state_raw_info log_prox_state_raw_info;
                  sns_memzero(&log_prox_state_raw_info, sizeof(log_prox_state_raw_info));

                  tmf8801_generateProxResponse(state, diag, timer_event.timeout_time,
                                                         this, &log_prox_state_raw_info, AMS_REPORT_STREAMING);
                }
            }
          }
        }
        else
        {
          AMS_DEV_INST_PRINTF(LOW, this
                           , "AMS: tmf8801_inst_notify_event: ignoring timer evt; timer_is_active=%d, S.I. > 0? _%d_"
                           , state->timer_is_active
                           , (state->sampling_intvl > 0)
                           );
        }
      }
      else
      {
        AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: ERROR decoding timer_data_stream event");
      }
      event = state->timer_data_stream->api->get_next_input(state->timer_data_stream);
    }
    if (updateAvailableFlags == AMS_CALIBRATION_DONE)
    {
      _tmf8801_check_and_remove_stream(this, AMS_STRM_IDX, state->timer_data_stream);
    }
    AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: no more timer_data_stream events");
  }

  // Turn COM port OFF
  state->scp_service->api->sns_scp_update_bus_power(state->com_port_info.port_handle,
                                                                           false);
  AMS_DEV_INST_PRINTF(LOW, this, "AMS: tmf8801_inst_notify_event: close COM PORT");
  return SNS_RC_SUCCESS;
}

static void _tmf8801_check_and_remove_stream(sns_sensor_instance *this,
    ams_mode_index oc_strm, sns_data_stream *stream)
{
  ams_instance_state *state;
  bool test_client_present = false;

  if (NULL == stream) {
    return;
  }
  state = (ams_instance_state *) this->state->state;

  if (state->prox_info.test_info[oc_strm].test_client_present)
  {
    state->prox_info.test_info[oc_strm].test_client_present = false;
    test_client_present = true;
  }
  if (test_client_present && state->client_present == 0)
  {
    sns_service_manager *service_mgr = this->cb->get_service_manager(this);
    sns_stream_service *stream_mgr = (sns_stream_service*)
                service_mgr->get_service(service_mgr, SNS_STREAM_SERVICE);
    stream_mgr->api->remove_stream(stream_mgr, stream);
    if (oc_strm == AMS_STRM_IDX) {
      state->timer_is_active = false;
      state->timer_data_stream = NULL;
    } else {
      state->interrupt_data_stream = NULL;
    }
  }
}

/**
 * Publishes a calibration event
 *
 * @param[i] instance  Instance reference
 * @param[i] uid       Sensor UID
 * @param[i] scale_result
 * @param[i] scale_size
 * @param[i] bias_result
 * @param[i] bias_size
 *
 * @return none
 */
void _tmf8801_send_cal_event(sns_sensor_instance *instance,
                                sns_sensor_uid *uid,
                                float *scale_result, uint8_t scale_size,
                                float *bias_result, uint8_t bias_size)
{
  uint8_t arr_index = 0;
  float scale_arr[20];
  float bias_arr[20];
  pb_float_arr_arg scale_arg;
  pb_float_arr_arg bias_arg;
  sns_cal_event cal_event = sns_cal_event_init_default;
  cal_event.status = SNS_STD_SENSOR_SAMPLE_STATUS_ACCURACY_HIGH;

  if (0 == scale_size && 0 == bias_size) {
    return;
  }

  if (scale_size > 0)
  {
    sns_memscpy (scale_arr, sizeof(float) * scale_size, scale_result, sizeof(float) * scale_size);
    scale_arg.arr = scale_arr;
    scale_arg.arr_len = scale_size;
    scale_arg.arr_index = &arr_index;

    cal_event.scale_factor.funcs.encode = &pb_encode_float_arr_cb;
    cal_event.scale_factor.arg = &scale_arg;
  }
  if (bias_size > 0)
  {
    sns_memscpy (bias_arr, sizeof(float) * bias_size, bias_result, sizeof(float) * bias_size);
    bias_arg.arr = bias_arr;
    bias_arg.arr_len = bias_size;
    bias_arg.arr_index = &arr_index;

    cal_event.bias.funcs.encode = &pb_encode_float_arr_cb;
    cal_event.bias.arg = &bias_arg;
  }

  pb_send_event(instance,
                sns_cal_event_fields,
                &cal_event,
                sns_get_system_time(),
                SNS_CAL_MSGID_SNS_CAL_EVENT,
                uid);

  AMS_DEV_INST_PRINTF(LOW, instance, "AMS: _tmf8801_send_cal_event: CALIBRATION COMPLETED\n");
}

/**
 * Runs a factory test
 * test event.
 *
 * @param[i] instance  Instance reference
 * @param[i] uid       Sensor UID
 *
 * @return none
 */
void _tmf8801_send_factory_test_event(sns_sensor_instance *instance,
                                  sns_sensor_uid *uid, bool test_result)
{
  uint8_t data[1] = {0};
  pb_buffer_arg buff_arg = (pb_buffer_arg)
      { .buf = &data, .buf_len = sizeof(data) };
  sns_physical_sensor_test_event test_event =
     sns_physical_sensor_test_event_init_default;
  test_event.test_passed = test_result;
  test_event.test_type = SNS_PHYSICAL_SENSOR_TEST_TYPE_FACTORY;
  test_event.test_data.funcs.encode = &pb_encode_string_cb;
  test_event.test_data.arg = &buff_arg;

  pb_send_event(instance,
                sns_physical_sensor_test_event_fields,
                &test_event,
                sns_get_system_time(),
                SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_EVENT,
                uid);
  return;
}

void _tmf8801_send_factory_hw_test_event(sns_sensor_instance *instance,
                                  sns_sensor_uid *uid, bool test_result)
{
  uint8_t data[1] = {0};
  pb_buffer_arg buff_arg = (pb_buffer_arg)
      { .buf = &data, .buf_len = sizeof(data) };
  sns_physical_sensor_test_event test_event =
     sns_physical_sensor_test_event_init_default;
  test_event.test_passed = test_result;
  test_event.test_type = SNS_PHYSICAL_SENSOR_TEST_TYPE_HW;
  test_event.test_data.funcs.encode = &pb_encode_string_cb;
  test_event.test_data.arg = &buff_arg;

  pb_send_event(instance,
                sns_physical_sensor_test_config_fields,
                &test_event,
                sns_get_system_time(),
                SNS_PHYSICAL_SENSOR_TEST_MSGID_SNS_PHYSICAL_SENSOR_TEST_EVENT,
                uid);
  return;
}

/**
 * Collects factory test results and submits for replying and/or updating
 * the cal data in the registry.
 *
 * @param[i] instance     Instance reference
 * @param[i] test_result  Test pass/fail result if already known, else NULL
 *
 * @return none
 */
static void _tmf8801_wrap_up_factory_test(sns_sensor_instance *this,
                                         bool *test_result)
{
  ams_instance_state *state = (ams_instance_state*)this->state->state;
  int32_t updateAvailableFlags = 0;
  bool lcl_test_result = 0;

  if(test_result)
  {
    AMS_DEV_INST_PRINTF(MED, this
                    , "AMS: _tmf8801_wrap_up_factory_test: enter: instance %p, *test_result=%d"
                    , this
                    , *test_result
                    );
  } else {
    AMS_DEV_INST_PRINTF(MED, this
                    , "AMS: _tmf8801_wrap_up_factory_test: enter: instance %p, no incoming test_result"
                    , this
                    );
  }

  if(!test_result)
  {
    test_result = &lcl_test_result;
    updateAvailableFlags = tmf8801_getUpdateResult(state->priv);
    AMS_DEV_INST_PRINTF(MED, this
                    , "AMS: _tmf8801_wrap_up_factory_test: got cal status=%d"
                    , updateAvailableFlags
                    );
    if((state->sensor_type_under_test == AMS_PROX_OC)
        || (state->sensor_type_under_test == AMS_PROX_STRM))
    {
      if (updateAvailableFlags == AMS_CALIBRATION_DONE){
        *test_result = true;
      }
    }
  }

  if(state->prox_info.test_info[AMS_OC_IDX].test_client_present)
  {
    if(state->prox_info.test_info[AMS_OC_IDX].test_type == SNS_PHYSICAL_SENSOR_TEST_TYPE_FACTORY)
    {

      AMS_DEV_INST_PRINTF(MED, this
                      , "AMS: _tmf8801_wrap_up_factory_test: PROX_OC, *test_result=%d"
                      , *test_result
                      );
      state->cal_version[TOF_FAC_CAL_VERSION]++;

      _tmf8801_send_cal_event (this, &state->prox_info.suid[AMS_OC_IDX],
                                state->instanceCalDataFloat.calData, TMF8801_CALIBRATION_DATA_SIZE, NULL, 0);
      _tmf8801_send_factory_test_event(this, &state->prox_info.suid[AMS_OC_IDX], *test_result);
      state->new_self_test_request = false;
    }
    return;
  }

  if(state->prox_info.test_info[AMS_STRM_IDX].test_client_present)
  {
    if(state->prox_info.test_info[AMS_STRM_IDX].test_type == SNS_PHYSICAL_SENSOR_TEST_TYPE_FACTORY)
    {

      AMS_DEV_INST_PRINTF(MED, this
                      , "AMS: _tmf8801_wrap_up_factory_test: PROX_STRM, *test_result=%d"
                      , *test_result
                      );
      state->cal_version[TOF_FAC_CAL_VERSION]++;
      _tmf8801_send_cal_event (this, &state->prox_info.suid[AMS_STRM_IDX],
                                state->instanceCalDataFloat.calData, TMF8801_CALIBRATION_DATA_SIZE, NULL, 0);

      _tmf8801_send_factory_test_event(this, &state->prox_info.suid[AMS_STRM_IDX], *test_result);
      state->new_self_test_request = false;
    }
    return;
  }
}

/** Public Data Definitions. */

sns_sensor_instance_api tmf8801_sensor_instance_api =
{
  .struct_len        = sizeof(sns_sensor_instance_api),
  .init              = &tmf8801_inst_init,
  .deinit            = &tmf8801_inst_deinit,
  .set_client_config = &tmf8801_inst_set_client_config,
  .notify_event      = &tmf8801_inst_notify_event
};
