#!/usr/bin/env python

# This script takes as input an Intel hex record file
# and embeds the records in a JSON file.
#
# ------------------------------------------------------------------------------
# Input file format
# ------------------------------------------------------------------------------
# :020000042000DA
# :100000006DC941853D15AA51F4D29EA8A7AC77E9E8
# .
# .
# .
# :00000001FF
#
# ------------------------------------------------------------------------------
# Output file format
# ------------------------------------------------------------------------------
# {
#   "config":{
#     "hw_platform": ["MTP", "Surf", "QRD", "HDK"],
#     "soc_id": ["321", "341"]
#   },
#   "tmf8801_0_firmware":{
#     "owner": "sns_tmf8801",
#     ".fw":{
#       "owner": "sns_tmf8801",
#       "record_count":{ "type": "int", "ver": "0",
#         "data": "616"
#       },
#       "rec_0":{ "type": "str", "ver": "0",
#         "data": "020000042000DA"
#       },
#       "rec_1":{ "type": "str", "ver": "0",
#         "data": "100000006DC941853D15AA51F4D29EA8A7AC77E9E8"
#       },
# .
# .
# .
#       "rec_615":{ "type": "str", "ver": "0",
#         "data": "00000001FF"
#       }
#     }
#   }
# }


import sys, argparse

class cfile(file):
    #subclass file to have a more convienient use of writeline
    def __init__(self, name, mode = 'r'):
        self = file.__init__(self, name, mode)

    def wl(self, string):
        self.writelines(string + '\n')
        return None

def file_len(ifp):
    lines = 0
    for line in ifp:
        lines += 1
    return lines
        
def write_json_header(ofp, record_count):
    ofp.wl('{');
    ofp.wl('  "config":{');
    ofp.wl('    "hw_platform": ["MTP", "Surf", "QRD", "HDK"],');
    ofp.wl('    "soc_id": ["321", "341"]');
    ofp.wl('  },');
    ofp.wl('  "tmf8801_0_firmware":{');
    ofp.wl('    "owner": "sns_tmf8801",');
    ofp.wl('    ".fw":{');
    ofp.wl('      "owner": "sns_tmf8801",');
    ofp.wl('      "record_count":{ "type": "int", "ver": "0",');
    ofp.wl('        "data": "{}"'.format(record_count));
    ofp.wl('      },');
    
def write_hex_record_to_json(ofp, recno, record, final):
    ofp.wl('      "rec_{}":'.format(recno) +'{ "type": "str", "ver": "0",');
    ofp.wl('        "data": "{}"'.format(record));
    if (final == True):
        ofp.wl('      }');
    else:
        ofp.wl('      },');
    
def write_json_footer(ofp):
    ofp.wl('    }');
    ofp.wl('  }');
    ofp.wl('}');

def print_usage():
    print ("hex2json usage:\n")
    print ("    hex2json.py -i <inputfile> [-o <outputfile>]\n")
    
def main():
    parser = argparse.ArgumentParser(description = "A tool to wrap firmware hex records in JSON")
    parser.add_argument("-i", "--inputfile", help = "Hex record input file", required = True, default = "")
    parser.add_argument("-o", "--outputfile", help = "JSON output file", required = False, default = "tmf8801_0_firmware.json")
    
    arguments = parser.parse_args()

    inputfilename = arguments.inputfile
    outputfilename = arguments.outputfile
    
    # Open the input file
    try:
        ifp = open(inputfilename, "r")
    except IOError:
        print ("Error: input file does not exist")
        sys.exit()

        
    line_count = file_len(ifp);
    if (line_count != 0):
        # Only need to open output file if there are records to process
        ofp = cfile(outputfilename, "w")

        print ("Input file contains {} records".format(line_count));
        print ("Writing JSON")
        
        # reset the input file pointer
        ifp.seek(0)

        # Process records
        rec_number = 0
        write_json_header(ofp, line_count)
        line = ifp.readline()
        while line:
            # Remove : from beginning of line, also leading and trailing whitespace
            trimmed_record_string = (line.strip().replace(":",""))
            if (rec_number == (line_count - 1)):
                write_hex_record_to_json(ofp, rec_number, trimmed_record_string, True)
            else:
                write_hex_record_to_json(ofp, rec_number, trimmed_record_string, False)
                
            rec_number += 1
            line = ifp.readline()

        write_json_footer(ofp)

        ifp.close();
        ofp.close();
        print ("Done")
    else:
        print("Nothing to do: input file is empty")


if __name__ == "__main__":
    main()

